"""This module provides proxy objects around objects from the common
workflow language reference implementation library cwltool. These proxies
adapt cwltool to Galaxy features and abstract the library away from the rest
of the framework.
"""

import copy
import json
import logging
import os
from abc import (
    ABCMeta,
    abstractmethod,
)
from typing import (
    Any,
    Dict,
    List,
    Optional,
    overload,
    TYPE_CHECKING,
    Union,
)
from uuid import (
    UUID,
    uuid4,
)

from typing_extensions import (
    Literal,
    TypedDict,
)

from galaxy.exceptions import MessageException
from galaxy.util import (
    listify,
    safe_makedirs,
)
from galaxy.util.bunch import Bunch
from .cwltool_deps import (
    beta_relaxed_fmt_check,
    ensure_cwltool_available,
    getdefault,
    normalizeFilesDirs,
    pathmapper,
    process,
    ref_resolver,
    relink_initialworkdir,
    RuntimeContext,
    sourceline,
    StdFsAccess,
    visit_class,
    yaml_no_ts,
)
from .representation import (
    field_to_field_type,
    FIELD_TYPE_REPRESENTATION,
    INPUT_TYPE,
    type_descriptions_for_field_types,
    USE_FIELD_TYPES,
    USE_STEP_PARAMETERS,
)
from .schema import (
    non_strict_non_validating_schema_loader,
    schema_loader,
)
from .util import SECONDARY_FILES_EXTRA_PREFIX

if TYPE_CHECKING:
    from .cwltool_deps import (
        CWLObjectType,
        JobsType,
        Process,
        workflow,
    )
    from .schema import RawProcessReference

log = logging.getLogger(__name__)

JOB_JSON_FILE = ".cwl_job.json"

DOCKER_REQUIREMENT = "DockerRequirement"
SUPPORTED_TOOL_REQUIREMENTS = [
    "CreateFileRequirement",
    "DockerRequirement",
    "EnvVarRequirement",
    "InitialWorkDirRequirement",
    "InlineJavascriptRequirement",
    "ResourceRequirement",
    "ShellCommandRequirement",
    "ScatterFeatureRequirement",
    "SchemaDefRequirement",
    "SubworkflowFeatureRequirement",
    "StepInputExpressionRequirement",
    "MultipleInputFeatureRequirement",
    "CredentialsRequirement",
]


SUPPORTED_WORKFLOW_REQUIREMENTS = SUPPORTED_TOOL_REQUIREMENTS + []

ToolStateType = Dict[str, Union[None, str, bool, Dict[str, str]]]


class InputInstanceDict(TypedDict, total=False):
    type: str
    name: str
    title: str
    label: str
    help: str
    optional: bool
    area: bool
    value: str
    collection_type: str


class InputInstanceArrayDict(TypedDict):
    type: str
    name: str
    title: str
    blocks: List[InputInstanceDict]


class ToolProxy(metaclass=ABCMeta):
    _class: str

    def __init__(
        self,
        tool: "Process",
        uuid: Union[UUID, str],
        raw_process_reference: Optional["RawProcessReference"] = None,
        tool_path: Optional[str] = None,
    ):
        self._tool = tool
        self.uuid = uuid
        self._tool_path = tool_path
        self._raw_process_reference = raw_process_reference
        # remove input parameter formats from CWL files so that cwltool
        # does not complain they are missing in the input data
        assert isinstance(self._tool.inputs_record_schema, dict)
        for input_field in self._tool.inputs_record_schema["fields"]:
            if "format" in input_field:
                del input_field["format"]

    def job_proxy(self, input_dict: Dict[str, Any], output_dict, job_directory: str = "."):
        """Build a cwltool.job.Job describing computation using a input_json
        Galaxy will generate mapping the Galaxy description of the inputs into
        a cwltool compatible variant.
        """
        return JobProxy(self, input_dict, output_dict, job_directory=job_directory)

    @property
    def id(self):
        raw_id = self._tool.tool.get("id", None)
        return raw_id

    def galaxy_id(self) -> str:
        raw_id = self.id
        tool_id = None
        # don't reduce "search.cwl#index" to search
        if raw_id:
            tool_id = os.path.basename(raw_id)
            # tool_id = os.path.splitext(os.path.basename(raw_id))[0]
        if not tool_id:
            return str(self.uuid)
        if tool_id.startswith("#"):
            tool_id = tool_id[1:]
        return tool_id

    @abstractmethod
    def input_fields(self) -> list:
        """Return InputInstance objects describing mapping to Galaxy inputs."""

    @abstractmethod
    def input_instances(self):
        """Return InputInstance objects describing mapping to Galaxy inputs."""

    @abstractmethod
    def output_instances(self) -> List["OutputInstance"]:
        """Return OutputInstance objects describing mapping to Galaxy inputs."""

    @abstractmethod
    def docker_identifier(self):
        """Return docker identifier for embedding in tool description."""

    @abstractmethod
    def description(self):
        """Return description to tool."""

    @abstractmethod
    def label(self):
        """Return label for tool."""

    def to_persistent_representation(self):
        """Return a JSON representation of this tool. Not for serialization
        over the wire, but serialization in a database."""
        persisted_obj = self._tool.tool
        persisted_obj["requirements"] = self.requirements
        if not persisted_obj.get("cwlVersion"):
            # This happens for any inline process, but getting it from metadata is correct for inline processes at least
            persisted_obj["cwlVersion"] = self._tool.metadata["cwlVersion"]
        return {
            "class": self._class,
            # Should maybe be yaml instead
            "raw_process_reference": persisted_obj,
            "uuid": self.uuid,
        }

    @staticmethod
    def from_persistent_representation(
        as_object: Dict[str, Any], strict_cwl_validation: bool = True, tool_directory: Optional[str] = None
    ) -> "ToolProxy":
        """Recover an object serialized with to_persistent_representation."""
        if "class" not in as_object:
            raise Exception("Failed to deserialize tool proxy from JSON object - no class found.")
        loaded_object = tool_proxy(
            tool_object=as_object["raw_process_reference"],
            strict_cwl_validation=strict_cwl_validation,
            tool_directory=tool_directory,
            uuid=as_object.get("uuid"),
        )
        return loaded_object

    @property
    def requirements(self) -> List:
        return getattr(self._tool, "requirements", [])

    def hints_or_requirements_of_class(self, class_name: str) -> List:
        reqs_and_hints = self.requirements + getattr(self._tool, "hints", [])
        return [hint for hint in reqs_and_hints if hint["class"] == class_name]

    def software_requirements(self) -> List:
        # Roughest imaginable pass at parsing requirements, really need to take in specs, handle
        # multiple versions, etc...
        requirements = []
        for hint in self.hints_or_requirements_of_class("SoftwareRequirement"):
            packages = hint.get("packages", [])
            for package in packages:
                versions = package.get("version", [])
                first_version = None if not versions else versions[0]
                requirements.append((package["package"], first_version))
        return requirements

    def resource_requirements(self) -> List:
        return self.hints_or_requirements_of_class("ResourceRequirement")

    def credentials_requirements(self) -> List:
        return self.hints_or_requirements_of_class("CredentialsRequirement")


class CommandLineToolProxy(ToolProxy):
    _class = "CommandLineTool"

    def description(self):
        # Don't use description - typically too verbose.
        return ""

    def doc(self):
        # TODO: parse multiple lines and merge - valid in cwl-1.1
        doc = self._tool.tool.get("doc")
        return doc

    def label(self):
        label = self._tool.tool.get("label")

        if label is not None:
            return label.partition(":")[0]  # return substring before ':'
        else:
            return ""

    def input_fields(self) -> list:
        input_records_schema = self._eval_schema(self._tool.inputs_record_schema)
        if input_records_schema["type"] != "record":
            raise Exception("Unhandled CWL tool input structure")

        # TODO: handle this somewhere else?
        # schemadef_req_tool_param
        rval = []
        for input in input_records_schema["fields"]:
            input_copy = copy.deepcopy(input)
            input_type = input.get("type")
            if isinstance(input_type, list) or isinstance(input_type, dict):
                rval.append(input_copy)
                continue

            if input_type in self._tool.schemaDefs:
                input_copy["type"] = self._tool.schemaDefs[input_type]

            rval.append(input_copy)
        return rval

    def _eval_schema(self, io_schema):
        schema_type = io_schema.get("type")
        if schema_type in self._tool.schemaDefs:
            io_schema = self._tool.schemaDefs[schema_type]
        return io_schema

    def input_instances(self):
        return [_outer_field_to_input_instance(_) for _ in self.input_fields()]

    def output_instances(self):
        outputs_schema = self._eval_schema(self._tool.outputs_record_schema)
        if outputs_schema["type"] != "record":
            raise Exception("Unhandled CWL tool output structure")

        return [_simple_field_to_output(output) for output in outputs_schema["fields"]]

    def docker_identifier(self):
        for hint in self.hints_or_requirements_of_class("DockerRequirement"):
            if "dockerImageId" in hint:
                return hint["dockerImageId"]
            else:
                return hint["dockerPull"]

        return None


class ExpressionToolProxy(CommandLineToolProxy):
    _class = "ExpressionTool"


class JobProxy:
    _is_command_line_job: bool

    def __init__(self, tool_proxy: ToolProxy, input_dict: Dict[str, Any], output_dict, job_directory: str):
        assert RuntimeContext is not None, "cwltool is not installed, cannot run CWL jobs"
        self._tool_proxy = tool_proxy
        self._input_dict = input_dict
        self._output_dict = output_dict
        self._job_directory = job_directory

        self._final_output: Optional[CWLObjectType] = None
        self._ok = True
        self._cwl_job: Optional[JobsType] = None

        self._normalize_job()

    def cwl_job(self):
        self._ensure_cwl_job_initialized()
        return self._cwl_job

    @property
    def is_command_line_job(self):
        self._ensure_cwl_job_initialized()
        return self._is_command_line_job

    def _ensure_cwl_job_initialized(self):
        if self._cwl_job is None:
            job_args = dict(
                basedir=self._job_directory,
                select_resources=self._select_resources,
                outdir=os.path.join(self._job_directory, "working"),
                tmpdir=os.path.join(self._job_directory, "cwltmp"),
                stagedir=os.path.join(self._job_directory, "cwlstagedir"),
                use_container=False,
                beta_relaxed_fmt_check=beta_relaxed_fmt_check,
            )

            runtimeContext = RuntimeContext(job_args)
            self._cwl_job = next(self._tool_proxy._tool.job(self._input_dict, self._output_callback, runtimeContext))
            self._is_command_line_job = hasattr(self._cwl_job, "command_line")

    def _normalize_job(self):
        # Somehow reuse whatever causes validate in cwltool... maybe?
        def pathToLoc(p):
            if "location" not in p and "path" in p:
                p["location"] = p["path"]
                del p["path"]

        runtime_context = RuntimeContext({})
        make_fs_access = getdefault(runtime_context.make_fs_access, StdFsAccess)
        fs_access = make_fs_access(runtime_context.basedir)
        process.fill_in_defaults(self._tool_proxy._tool.tool["inputs"], self._input_dict, fs_access)
        visit_class(self._input_dict, ("File", "Directory"), pathToLoc)
        # TODO: Why doesn't fillInDefault fill in locations instead of paths?
        normalizeFilesDirs(self._input_dict)
        # TODO: validate like cwltool process _init_job.
        #    validate.validate_ex(self.names.get_name("input_record_schema", ""), builder.job,
        #                         strict=False, logger=_logger_validation_warnings)

    def rewrite_inputs_for_staging(self):
        if hasattr(self._cwl_job, "pathmapper"):
            pass
            # DO SOMETHING LIKE THE FOLLOWING?
            # path_rewrites = {}
            # for f, p in self._cwl_job.pathmapper.items():
            #     if not p.staged:
            #         continue
            #     if p.type in ("File", "Directory"):
            #         path_rewrites[p.resolved] = p.target
            # for key, value in self._input_dict.items():
            #     if key in path_rewrites:
            #         self._input_dict[key]["location"] = path_rewrites[value]
        else:
            stagedir = os.path.join(self._job_directory, "cwlstagedir")
            safe_makedirs(stagedir)

            def stage_recursive(value):
                is_list = isinstance(value, list)
                is_dict = isinstance(value, dict)
                log.info(f"handling value {value}, is_list {is_list}, is_dict {is_dict}")
                if is_list:
                    for val in value:
                        stage_recursive(val)
                elif is_dict:
                    if "location" in value and "basename" in value:
                        location = value["location"]
                        basename = value["basename"]
                        if not location.endswith(basename):  # TODO: sep()[-1]
                            staged_loc = os.path.join(stagedir, basename)
                            if not os.path.exists(staged_loc):
                                os.symlink(location, staged_loc)
                            value["location"] = staged_loc
                    for dict_value in value.values():
                        stage_recursive(dict_value)
                else:
                    log.info("skipping simple value...")

            stage_recursive(self._input_dict)

    def _select_resources(self, request, runtime_context=None):
        new_request = request.copy()
        new_request["cores"] = "$GALAXY_SLOTS"
        return new_request

    @property
    def command_line(self):
        if self.is_command_line_job:
            return self.cwl_job().command_line
        else:
            return ["true"]

    @property
    def stdin(self):
        if self.is_command_line_job:
            return self.cwl_job().stdin
        else:
            return None

    @property
    def stdout(self):
        if self.is_command_line_job:
            return self.cwl_job().stdout
        else:
            return None

    @property
    def stderr(self):
        if self.is_command_line_job:
            return self.cwl_job().stderr
        else:
            return None

    @property
    def environment(self):
        if self.is_command_line_job:
            return self.cwl_job().environment
        else:
            return {}

    @property
    def generate_files(self):
        if self.is_command_line_job:
            return self.cwl_job().generatefiles
        else:
            return {}

    def _output_callback(self, out: Optional["CWLObjectType"], process_status: str):
        self._process_status = process_status
        if process_status == "success":
            self._final_output = out
        else:
            self._ok = False

        log.info(f"Output are {out}, status is {process_status}")

    def collect_outputs(self, tool_working_directory: str, rcode: int):
        if not self.is_command_line_job:
            cwl_job = self.cwl_job()
            if RuntimeContext is not None:
                cwl_job.run(RuntimeContext({}))
            else:
                cwl_job.run()
            if not self._ok:
                raise Exception(f"Final process state not ok, [{self._process_status}]")
            return self._final_output
        else:
            return self.cwl_job().collect_outputs(tool_working_directory, rcode)

    def save_job(self) -> None:
        job_file = self._job_file(self._job_directory)
        job_objects = {
            # "tool_path": os.path.abspath(self._tool_proxy._tool_path),
            "tool_representation": self._tool_proxy.to_persistent_representation(),
            "job_inputs": self._input_dict,
            "output_dict": self._output_dict,
        }
        json.dump(job_objects, open(job_file, "w"))

    def _output_extra_files_dir(self, output_name):
        output_id = self.output_id(output_name)
        return os.path.join(self._job_directory, "outputs", f"dataset_{output_id}_files")

    def output_id(self, output_name):
        output_id = self._output_dict[output_name]["id"]
        return output_id

    def output_path(self, output_name):
        output_id = self._output_dict[output_name]["path"]
        return output_id

    def output_directory_contents_dir(self, output_name, create=False):
        extra_files_dir = self._output_extra_files_dir(output_name)
        return extra_files_dir

    def output_secondary_files_dir(self, output_name, create=False):
        extra_files_dir = self._output_extra_files_dir(output_name)
        secondary_files_dir = os.path.join(extra_files_dir, SECONDARY_FILES_EXTRA_PREFIX)
        if create and not os.path.exists(secondary_files_dir):
            safe_makedirs(secondary_files_dir)
        return secondary_files_dir

    def stage_files(self):
        cwl_job = self.cwl_job()

        def stageFunc(resolved_path, target_path):
            log.info(f"resolving {resolved_path} to {target_path}")
            try:
                os.symlink(resolved_path, target_path)
            except OSError:
                pass

        if hasattr(cwl_job, "pathmapper"):
            process.stage_files(cwl_job.pathmapper, stageFunc, ignore_writable=True, symlink=False)

        if hasattr(cwl_job, "generatefiles"):
            outdir = os.path.join(self._job_directory, "working")
            # TODO: Why doesn't cwl_job.generatemapper work?
            generate_mapper = pathmapper.PathMapper(
                cwl_job.generatefiles["listing"], outdir, outdir, separateDirs=False
            )
            # TODO: figure out what inplace_update should be.
            inplace_update = cwl_job.inplace_update
            process.stage_files(generate_mapper, stageFunc, ignore_writable=inplace_update, symlink=False)
            relink_initialworkdir(generate_mapper, outdir, outdir, inplace_update=inplace_update)
        # else: expression tools do not have a path mapper.

    @staticmethod
    def _job_file(job_directory):
        return os.path.join(job_directory, JOB_JSON_FILE)


class WorkflowProxy:
    def __init__(self, workflow: "workflow.Workflow", workflow_path: Optional[str] = None):
        self._workflow = workflow
        self._workflow_path = workflow_path
        self._step_proxies: Optional[List[Union[SubworkflowStepProxy, ToolStepProxy]]] = None

    @property
    def cwl_id(self):
        return self._workflow.tool["id"]

    def get_outputs_for_label(self, label):
        outputs = []
        for output in self._workflow.tool["outputs"]:
            step, output_name = split_step_references(
                output["outputSource"],
                multiple=False,
                workflow_id=self.cwl_id,
            )
            if step == label:
                output_id = output["id"]
                if "#" not in self.cwl_id:
                    _, output_label = output_id.rsplit("#", 1)
                else:
                    _, output_label = output_id.rsplit("/", 1)

                outputs.append(
                    {
                        "output_name": output_name,
                        "label": output_label,
                    }
                )
        return outputs

    def tool_reference_proxies(self):
        """Fetch tool source definitions for all referenced tools."""
        references: List[ToolProxy] = []
        for step in self.step_proxies():
            references.extend(step.tool_reference_proxies())
        return references

    def step_proxies(self):
        if self._step_proxies is None:
            proxies = []
            num_input_steps = len(self._workflow.tool["inputs"])
            for i, step in enumerate(self._workflow.steps):
                proxies.append(build_step_proxy(self, step, i + num_input_steps))
            self._step_proxies = proxies
        return self._step_proxies

    @property
    def runnables(self):
        runnables = []
        for step in self._workflow.steps:
            if "run" in step.tool:
                runnables.append(step.tool["run"])
        return runnables

    def cwl_ids_to_index(self, step_proxies):
        index = 0
        cwl_ids_to_index = {}
        for input_dict in self._workflow.tool["inputs"]:
            cwl_ids_to_index[input_dict["id"]] = index
            index += 1

        for step_proxy in step_proxies:
            cwl_ids_to_index[step_proxy.cwl_id] = index
            index += 1

        return cwl_ids_to_index

    @property
    def output_labels(self):
        return [self.jsonld_id_to_label(o["id"]) for o in self._workflow.tool["outputs"]]

    def input_connections_by_step(self, step_proxies):
        cwl_ids_to_index = self.cwl_ids_to_index(step_proxies)
        input_connections_by_step = []
        for step_proxy in step_proxies:
            input_connections_step: Dict[str, List[Dict[str, str]]] = {}
            for input_proxy in step_proxy.input_proxies:
                cwl_source_id = input_proxy.cwl_source_id
                input_name = input_proxy.input_name
                # Consider only allow multiple if MultipleInputFeatureRequirement is enabled
                for output_step_name, output_name in split_step_references(cwl_source_id, workflow_id=self.cwl_id):
                    if "#" in self.cwl_id:
                        sep_on = "/"
                    else:
                        sep_on = "#"
                    output_step_id = self.cwl_id + sep_on + output_step_name

                    if output_step_id not in cwl_ids_to_index:
                        template = "Output [%s] does not appear in ID-to-index map [%s]."
                        msg = template % (output_step_id, cwl_ids_to_index.keys())
                        raise AssertionError(msg)

                    if input_name not in input_connections_step:
                        input_connections_step[input_name] = []

                    input_connections_step[input_name].append(
                        {"id": cwl_ids_to_index[output_step_id], "output_name": output_name, "input_type": "dataset"}
                    )

            input_connections_by_step.append(input_connections_step)

        return input_connections_by_step

    def to_dict(self):
        name = os.path.basename(
            self._workflow.tool.get("label") or self._workflow_path or "TODO - derive a name from ID"
        )
        steps = {}

        step_proxies = self.step_proxies()
        input_connections_by_step = self.input_connections_by_step(step_proxies)
        index = 0
        for i, input_dict in enumerate(self._workflow.tool["inputs"]):
            steps[index] = self.cwl_input_to_galaxy_step(input_dict, i)
            index += 1

        for i, step_proxy in enumerate(step_proxies):
            input_connections = input_connections_by_step[i]
            steps[index] = step_proxy.to_dict(input_connections)
            index += 1

        return {
            "name": name,
            "steps": steps,
            "annotation": self.cwl_object_to_annotation(self._workflow.tool),
        }

    def find_inputs_step_index(self, label):
        for i, input in enumerate(self._workflow.tool["inputs"]):
            if self.jsonld_id_to_label(input["id"]) == label:
                return i

        raise Exception(f"Failed to find index for label {label}")

    def jsonld_id_to_label(self, id):
        if "#" in self.cwl_id:
            return id.rsplit("/", 1)[-1]
        else:
            return id.rsplit("#", 1)[-1]

    def cwl_input_to_galaxy_step(self, input, i):
        input_type = input["type"]
        label = self.jsonld_id_to_label(input["id"])
        input_as_dict = {
            "id": i,
            "label": label,
            "position": {"left": 0, "top": 0},
            "annotation": self.cwl_object_to_annotation(input),
            "input_connections": {},  # Should the Galaxy API really require this? - Seems to.
            "workflow_outputs": self.get_outputs_for_label(label),
        }
        if input_type == "File" and "default" not in input:
            input_as_dict["type"] = "data_input"
        elif isinstance(input_type, dict) and input_type.get("type") == "array":
            input_as_dict["type"] = "data_collection_input"
            input_as_dict["collection_type"] = "list"
        elif isinstance(input_type, dict) and input_type.get("type") == "record":
            input_as_dict["type"] = "data_collection_input"
            input_as_dict["collection_type"] = "record"
        else:
            if USE_STEP_PARAMETERS:
                input_as_dict["type"] = "parameter_input"
                # TODO: dispatch on actual type so this doesn't always need
                # to be field - simpler types could be simpler inputs.
                tool_state: ToolStateType = {}
                tool_state["parameter_type"] = "field"
                default_set = "default" in input
                default_value = input.get("default")
                optional = default_set
                if isinstance(input_type, list) and "null" in input_type:
                    optional = True
                if not optional and isinstance(input_type, dict) and "type" in input_type:
                    raise ValueError("'type' detected in non-optional input dictionary.")
                if default_set:
                    tool_state["default"] = {"src": "json", "value": default_value}
                tool_state["optional"] = optional
                input_as_dict["tool_state"] = tool_state
            else:
                input_as_dict["type"] = "data_input"
                # TODO: format = expression.json

        return input_as_dict

    def cwl_object_to_annotation(self, cwl_obj):
        return cwl_obj.get("doc", None)


def tool_proxy(
    tool_path: Optional[str] = None,
    tool_object=None,
    strict_cwl_validation: bool = True,
    tool_directory: Optional[str] = None,
    uuid: Optional[Union[UUID, str]] = None,
) -> ToolProxy:
    """Provide a proxy object to cwltool data structures to just
    grab relevant data.
    """
    ensure_cwltool_available()
    return _to_cwl_tool_object(
        tool_path=tool_path,
        tool_object=tool_object,
        strict_cwl_validation=strict_cwl_validation,
        tool_directory=tool_directory,
        uuid=uuid,
    )


def tool_proxy_from_persistent_representation(
    persisted_tool: Dict[str, Any], strict_cwl_validation: bool = True, tool_directory: Optional[str] = None
) -> ToolProxy:
    """Load a ToolProxy from a previously persisted representation."""
    ensure_cwltool_available()
    return ToolProxy.from_persistent_representation(
        persisted_tool, strict_cwl_validation=strict_cwl_validation, tool_directory=tool_directory
    )


def workflow_proxy(workflow_path: str, strict_cwl_validation: bool = True) -> WorkflowProxy:
    ensure_cwltool_available()
    return _to_cwl_workflow_object(workflow_path, strict_cwl_validation=strict_cwl_validation)


def load_job_proxy(job_directory: str, strict_cwl_validation: bool = True) -> JobProxy:
    ensure_cwltool_available()
    job_objects_path = os.path.join(job_directory, JOB_JSON_FILE)
    job_objects = json.load(open(job_objects_path))
    job_inputs = job_objects["job_inputs"]
    output_dict = job_objects["output_dict"]
    persisted_tool = job_objects["tool_representation"]
    cwl_tool = tool_proxy_from_persistent_representation(
        persisted_tool=persisted_tool, strict_cwl_validation=strict_cwl_validation
    )
    return cwl_tool.job_proxy(job_inputs, output_dict, job_directory=job_directory)


def _to_cwl_tool_object(
    tool_path: Optional[str] = None,
    tool_object=None,
    strict_cwl_validation: bool = False,
    tool_directory: Optional[str] = None,
    uuid: Optional[Union[UUID, str]] = None,
) -> ToolProxy:
    if uuid is None:
        uuid = str(uuid4())
    schema_loader = _schema_loader(strict_cwl_validation)
    if tool_path is not None:
        assert tool_object is None

        raw_process_reference = schema_loader.raw_process_reference(tool_path)
        cwl_tool = schema_loader.tool(
            raw_process_reference=raw_process_reference,
        )
    elif tool_object is not None:
        # Allow loading tools from YAML...
        as_str = json.dumps(tool_object)
        tool_object = yaml_no_ts().load(as_str)
        path = tool_directory
        if path is None:
            path = os.getcwd()
        uri = f"{ref_resolver.file_uri(path)}/"
        sourceline.add_lc_filename(tool_object, uri)
        raw_process_reference = schema_loader.raw_process_reference_for_object(tool_object, uri=uri)
        cwl_tool = schema_loader.tool(
            raw_process_reference=raw_process_reference,
        )
    else:
        raise ValueError("Either tool_path or tool_object should be defined")

    if isinstance(cwl_tool, int):
        raise Exception("Failed to load tool.")

    raw_tool = cwl_tool.tool
    # Apply Galaxy hacks to CWL tool representation to bridge semantic differences
    # between Galaxy and cwltool.
    _hack_cwl_requirements(cwl_tool)
    check_requirements(raw_tool)
    return _cwl_tool_object_to_proxy(cwl_tool, uuid, raw_process_reference=raw_process_reference, tool_path=tool_path)


def _cwl_tool_object_to_proxy(
    cwl_tool: "Process",
    uuid: Union[UUID, str],
    raw_process_reference: Optional["RawProcessReference"] = None,
    tool_path: Optional[str] = None,
) -> ToolProxy:
    raw_tool = cwl_tool.tool
    if "class" not in raw_tool:
        raise Exception("File does not declare a class, not a valid Draft 3+ CWL tool.")

    process_class = raw_tool["class"]
    if process_class == "CommandLineTool":
        proxy_class = CommandLineToolProxy
    elif process_class == "ExpressionTool":
        proxy_class = ExpressionToolProxy
    else:
        raise Exception("File not a CWL CommandLineTool.")
    top_level_object = tool_path is not None
    if top_level_object and ("cwlVersion" not in raw_tool):
        raise Exception("File does not declare a CWL version, pre-draft 3 CWL tools are not supported.")

    return proxy_class(cwl_tool, uuid, raw_process_reference, tool_path)


def _to_cwl_workflow_object(workflow_path: str, strict_cwl_validation: bool = True) -> WorkflowProxy:
    cwl_workflow = _schema_loader(strict_cwl_validation).tool(path=workflow_path)
    raw_workflow = cwl_workflow.tool
    check_requirements(raw_workflow, tool=False)
    return WorkflowProxy(cwl_workflow, workflow_path)


def _schema_loader(strict_cwl_validation: bool):
    return schema_loader if strict_cwl_validation else non_strict_non_validating_schema_loader


def _hack_cwl_requirements(cwl_tool):
    move_to_hints: List[int] = []
    for i, requirement in enumerate(cwl_tool.requirements):
        if requirement["class"] == DOCKER_REQUIREMENT:
            move_to_hints.insert(0, i)

    for i in move_to_hints:
        del cwl_tool.requirements[i]
        cwl_tool.hints.append(requirement)


def check_requirements(rec, tool=True):
    if isinstance(rec, dict):
        if "requirements" in rec:
            for r in rec["requirements"]:
                if tool:
                    possible = SUPPORTED_TOOL_REQUIREMENTS
                else:
                    possible = SUPPORTED_WORKFLOW_REQUIREMENTS
                if r["class"] not in possible:
                    raise Exception(f"Unsupported requirement {r['class']}")
        for d in rec:
            check_requirements(rec[d], tool=tool)
    if isinstance(rec, list):
        for d in rec:
            check_requirements(d, tool=tool)


def split_step_references(step_references, workflow_id=None, multiple=True):
    """Split a CWL step input or output reference into step id and name."""
    # Trim off the workflow id part of the reference.
    step_references = listify(step_references)
    split_references = []

    for step_reference in step_references:
        if workflow_id is None:
            # This path works fine for some simple workflows - but not so much
            # for subworkflows (maybe same for $graph workflows?)
            assert "#" in step_reference
            _, step_reference = step_reference.split("#", 1)
        else:
            if "#" in workflow_id:
                sep_on = "/"
            else:
                sep_on = "#"
            expected_prefix = workflow_id + sep_on
            if not step_reference.startswith(expected_prefix):
                raise AssertionError(f"step_reference [{step_reference}] doesn't start with {expected_prefix}")
            step_reference = step_reference[len(expected_prefix) :]

        # Now just grab the step name and input/output name.
        assert "#" not in step_reference
        if "/" in step_reference:
            step_name, io_name = step_reference.split("/", 1)
        else:
            # Referencing an input, not a step.
            # In Galaxy workflows input steps have an implicit output named
            # "output" for consistency with tools - in cwl land
            # just the input name is referenced.
            step_name = step_reference
            io_name = "output"
        split_references.append((step_name, io_name))

    if multiple:
        return split_references
    else:
        assert len(split_references) == 1, split_references
        return split_references[0]


def build_step_proxy(workflow_proxy: WorkflowProxy, step: "workflow.WorkflowStep", index: int):
    step_type = step.embedded_tool.tool["class"]
    if step_type == "Workflow":
        return SubworkflowStepProxy(workflow_proxy, step, index)
    else:
        return ToolStepProxy(workflow_proxy, step, index)


class InputProxy:
    def __init__(self, step_proxy, cwl_input):
        self._cwl_input = cwl_input
        self.step_proxy = step_proxy
        self.workflow_proxy = step_proxy._workflow_proxy

        cwl_input_id = cwl_input["id"]
        cwl_source_id = cwl_input.get("source", None)
        if cwl_source_id is None:
            if "valueFrom" not in cwl_input and "default" not in cwl_input:
                msg = f"Workflow step input must define a source, a valueFrom, or a default value. Obtained [{cwl_input}]."
                raise MessageException(msg)

        assert cwl_input_id
        step_name, input_name = split_step_references(
            cwl_input_id, multiple=False, workflow_id=step_proxy.cwl_workflow_id
        )
        self.step_name = step_name
        self.input_name = input_name

        self.cwl_input_id = cwl_input_id
        self.cwl_source_id = cwl_source_id

        scatter_inputs = [
            split_step_references(i, multiple=False, workflow_id=step_proxy.cwl_workflow_id)[1]
            for i in listify(step_proxy._step.tool.get("scatter", []))
        ]
        scatter = self.input_name in scatter_inputs
        self.scatter = scatter

    def to_dict(self):
        as_dict = {
            "name": self.input_name,
        }
        if "linkMerge" in self._cwl_input:
            as_dict["merge_type"] = self._cwl_input["linkMerge"]
        if "scatterMethod" in self.step_proxy._step.tool:
            as_dict["scatter_type"] = self.step_proxy._step.tool.get("scatterMethod", "dotproduct")
        else:
            as_dict["scatter_type"] = "dotproduct" if self.scatter else "disabled"
        if "valueFrom" in self._cwl_input:
            # TODO: Add a table for expressions - mark the type as CWL 1.0 JavaScript.
            as_dict["value_from"] = self._cwl_input["valueFrom"]
        if "default" in self._cwl_input:
            as_dict["default"] = self._cwl_input["default"]
        return as_dict


class BaseStepProxy:
    def __init__(self, workflow_proxy: WorkflowProxy, step: "workflow.WorkflowStep", index: int):
        self._workflow_proxy = workflow_proxy
        self._step = step
        self._index = index
        self._uuid = str(uuid4())
        cwl_inputs = self._step.tool["inputs"]
        self.input_proxies = [InputProxy(self, cwl_input) for cwl_input in cwl_inputs]

    @property
    def step_class(self):
        return self.cwl_tool_object.tool["class"]

    @property
    def cwl_id(self):
        return self._step.id

    @property
    def cwl_workflow_id(self):
        return self._workflow_proxy.cwl_id

    @property
    def requirements(self):
        return self._step.requirements

    @property
    def hints(self):
        return self._step.hints

    @property
    def label(self):
        label = self._workflow_proxy.jsonld_id_to_label(self._step.id)
        return label

    def galaxy_workflow_outputs_list(self):
        return self._workflow_proxy.get_outputs_for_label(self.label)

    @property
    def cwl_tool_object(self):
        return self._step.embedded_tool

    def inputs_to_dicts(self):
        inputs_as_dicts = []
        for input_proxy in self.input_proxies:
            inputs_as_dicts.append(input_proxy.to_dict())
        return inputs_as_dicts


class ToolStepProxy(BaseStepProxy):
    def __init__(self, workflow_proxy, step, index):
        super().__init__(workflow_proxy, step, index)
        self._tool_proxy = None

    @property
    def tool_proxy(self):
        # Needs to be cached so UUID that is loaded matches UUID generated with to_dict.
        if self._tool_proxy is None:
            self._tool_proxy = _cwl_tool_object_to_proxy(self.cwl_tool_object, uuid=str(uuid4()))
        return self._tool_proxy

    def tool_reference_proxies(self):
        # Return a list so we can handle subworkflows recursively.
        return [self.tool_proxy]

    def to_dict(self, input_connections):
        # We need to stub out null entries for things getting replaced by
        # connections. This doesn't seem ideal - consider just making Galaxy
        # handle this.
        tool_state: ToolStateType = {}
        for input_name in input_connections.keys():
            tool_state[input_name] = None

        outputs = self.galaxy_workflow_outputs_list()
        when_expression = self._step.tool.get("when")
        rval = {
            "id": self._index,
            "tool_uuid": self.tool_proxy.uuid,  # TODO: make sure this is respected...
            "label": self.label,
            "position": {"left": 0, "top": 0},
            "type": "tool",
            "annotation": self._workflow_proxy.cwl_object_to_annotation(self._step.tool),
            "input_connections": input_connections,
            "inputs": self.inputs_to_dicts(),
            "workflow_outputs": outputs,
        }
        if when_expression:
            rval["when"] = when_expression
        return rval


class SubworkflowStepProxy(BaseStepProxy):
    def __init__(self, workflow_proxy, step, index):
        super().__init__(workflow_proxy, step, index)
        self._subworkflow_proxy = None

    def to_dict(self, input_connections):
        outputs = self.galaxy_workflow_outputs_list()
        for key, input_connection_list in input_connections.items():
            input_subworkflow_step_id = self.subworkflow_proxy.find_inputs_step_index(key)
            for input_connection in input_connection_list:
                input_connection["input_subworkflow_step_id"] = input_subworkflow_step_id

        return {
            "id": self._index,
            "label": self.label,
            "position": {"left": 0, "top": 0},
            "type": "subworkflow",
            "subworkflow": self.subworkflow_proxy.to_dict(),
            "annotation": self.subworkflow_proxy.cwl_object_to_annotation(self._step.tool),
            "input_connections": input_connections,
            "inputs": self.inputs_to_dicts(),
            "workflow_outputs": outputs,
        }

    def tool_reference_proxies(self):
        return self.subworkflow_proxy.tool_reference_proxies()

    @property
    def subworkflow_proxy(self):
        if self._subworkflow_proxy is None:
            self._subworkflow_proxy = WorkflowProxy(self.cwl_tool_object)
        return self._subworkflow_proxy


def _outer_field_to_input_instance(field):
    field_type = field_to_field_type(field)  # Must be a list if in here?
    if not isinstance(field_type, list):
        field_type = [field_type]

    name, label, description = _field_metadata(field)

    case_name = "_cwl__type_"
    case_label = f"Specify Parameter {label} As"

    def value_input(type_description):
        value_name = "_cwl__value_"
        value_label = label
        value_description = description
        return InputInstance(
            value_name,
            value_label,
            value_description,
            input_type=type_description.galaxy_param_type,
            collection_type=type_description.collection_type,
        )

    select_options = []
    case_options = []
    type_descriptions = type_descriptions_for_field_types(field_type)
    for type_description in type_descriptions:
        select_options.append({"value": type_description.name, "label": type_description.label})
        input_instances = []
        if type_description.uses_param:
            input_instances.append(value_input(type_description))
        case_options.append((type_description.name, input_instances))

    # If there is more than one way to represent this parameter - produce a conditional
    # requesting user to ask for what form they want to submit the data in, else just map
    # a simple Galaxy parameter.
    if len(case_options) > 1 and not USE_FIELD_TYPES:
        case_input = SelectInputInstance(
            name=case_name,
            label=case_label,
            description=False,
            options=select_options,
        )

        return ConditionalInstance(name, case_input, case_options)
    else:
        if len(case_options) > 1:
            only_type_description = FIELD_TYPE_REPRESENTATION
        else:
            only_type_description = type_descriptions[0]

        return InputInstance(
            name,
            label,
            description,
            input_type=only_type_description.galaxy_param_type,
            collection_type=only_type_description.collection_type,
        )

    # Older array to repeat handling, now we are just representing arrays as
    # dataset collections - we should offer a blended approach in the future.
    # if field_type in simple_map_type_map.keys():
    #     input_type = simple_map_type_map[field_type]
    #     return {"input_type": input_type, "array": False}
    # elif field_type == "array":
    #     if isinstance(field["type"], dict):
    #         array_type = field["type"]["items"]
    #     else:
    #         array_type = field["items"]
    #     if array_type in simple_map_type_map.keys():
    #         input_type = simple_map_type_map[array_type]
    #     return {"input_type": input_type, "array": True}
    # else:
    #     raise Exception("Unhandled simple field type encountered - [%s]." % field_type)


def _field_metadata(field):
    name = field["name"]
    label = field.get("label", None)
    description = field.get("doc", None)
    return name, label, description


def _simple_field_to_output(field):
    name = field["name"]
    output_data_class = field["type"]
    output_instance = OutputInstance(name, output_data_type=output_data_class, output_type=OUTPUT_TYPE.GLOB)
    return output_instance


class ConditionalInstance:
    def __init__(self, name, case, whens):
        self.input_type = INPUT_TYPE.CONDITIONAL
        self.name = name
        self.case = case
        self.whens = whens

    def to_dict(self):
        as_dict = dict(
            name=self.name,
            type=INPUT_TYPE.CONDITIONAL,
            test=self.case.to_dict(),
            when={},
        )
        for value, block in self.whens:
            as_dict["when"][value] = [i.to_dict() for i in block]

        return as_dict


class SelectInputInstance:
    def __init__(self, name, label, description, options):
        self.input_type = INPUT_TYPE.SELECT
        self.name = name
        self.label = label
        self.description = description
        self.options = options

    def to_dict(self):
        # TODO: serialize options...
        as_dict = dict(
            name=self.name,
            label=self.label or self.name,
            help=self.description,
            type=self.input_type,
            options=self.options,
        )
        return as_dict


class InputInstance:
    def __init__(self, name, label, description, input_type, array=False, area=False, collection_type=None):
        self.input_type = input_type
        self.collection_type = collection_type
        self.name = name
        self.label = label
        self.description = description
        self.required = True
        self.array = array
        self.area = area

    @overload
    def to_dict(self, itemwise: Literal[False]) -> InputInstanceDict: ...

    @overload
    def to_dict(self, itemwise: Literal[True]) -> Union[InputInstanceDict, InputInstanceArrayDict]: ...

    def to_dict(self, itemwise: bool = True) -> Union[InputInstanceDict, InputInstanceArrayDict]:
        if itemwise and self.array:
            return InputInstanceArrayDict(
                type="repeat", name=f"{self.name}_repeat", title=f"{self.name}", blocks=[self.to_dict(itemwise=False)]
            )
        else:
            as_dict = InputInstanceDict(
                name=self.name,
                label=self.label or self.name,
                help=self.description,
                type=self.input_type,
                optional=not self.required,
            )
            if self.area:
                as_dict["area"] = True

            if self.input_type == INPUT_TYPE.INTEGER:
                as_dict["value"] = "0"
            if self.input_type == INPUT_TYPE.FLOAT:
                as_dict["value"] = "0.0"
            elif self.input_type == INPUT_TYPE.DATA_COLLECTION:
                as_dict["collection_type"] = self.collection_type
            return as_dict


OUTPUT_TYPE = Bunch(
    GLOB="glob",
    STDOUT="stdout",
)


# TODO: Different subclasses - this is representing different types of things.
class OutputInstance:
    def __init__(self, name: str, output_data_type, output_type, path=None, fields=None):
        self.name = name
        self.output_data_type = output_data_type
        self.output_type = output_type
        self.path = path
        self.fields = fields


__all__ = (
    "tool_proxy",
    "load_job_proxy",
)



"""Types used by interactor and test case processor."""

from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
)

from typing_extensions import (
    NotRequired,
    TypedDict,
)

from galaxy.tool_util.parser.interface import (
    AssertionList,
    TestSourceTestOutputColllection,
    ToolSourceTestOutputs,
)

# legacy inputs for working with POST /api/tools
# + inputs that have been processed with parse.py and expanded out
ExpandedToolInputs = Dict[str, Any]
# + ExpandedToolInputs where any model objects have been json-ified with to_dict()
ExpandedToolInputsJsonified = Dict[str, Any]

# modern inputs for working with POST /api/jobs*
RawTestToolRequest = Dict[str, Any]

ExtraFileInfoDictT = Dict[str, Any]
RequiredFileTuple = Tuple[str, ExtraFileInfoDictT]
RequiredFilesT = List[RequiredFileTuple]
RequiredDataTablesT = List[str]
RequiredLocFileT = List[str]


class ToolTestDescriptionDict(TypedDict):
    tool_id: str
    tool_version: Optional[str]
    name: str
    test_index: int
    inputs: ExpandedToolInputsJsonified
    request: NotRequired[Optional[Dict[str, Any]]]
    request_schema: NotRequired[Optional[Dict[str, Any]]]
    outputs: ToolSourceTestOutputs
    output_collections: List[TestSourceTestOutputColllection]
    stdout: Optional[AssertionList]
    stderr: Optional[AssertionList]
    expect_exit_code: Optional[int]
    expect_failure: bool
    expect_test_failure: bool
    num_outputs: Optional[int]
    command_line: Optional[AssertionList]
    command_version: Optional[AssertionList]
    required_files: List[Any]
    required_data_tables: List[Any]
    required_loc_files: List[str]
    error: bool
    exception: Optional[str]
    maxseconds: NotRequired[Optional[int]]



import io

try:
    import h5py
except ImportError:
    h5py = None

from ._types import (
    Annotated,
    AssertionParameter,
    OutputBytes,
)

IMPORT_MISSING_MESSAGE = "h5 assertion requires unavailable optional dependency h5py"

Key = Annotated[str, AssertionParameter("HDF5 attribute to check value of.")]
Value = Annotated[str, AssertionParameter("Expected value of HDF5 attribute to check.")]
Keys = Annotated[str, AssertionParameter("HDF5 attributes to check value of as a comma-separated string.")]


def _assert_h5py():
    if h5py is None:
        raise Exception(IMPORT_MISSING_MESSAGE)


def assert_has_h5_attribute(output_bytes: OutputBytes, key: Key, value: Value) -> None:
    """Asserts HDF5 output contains the specified ``value`` for an attribute (``key``), e.g.

    ```xml
    <has_h5_attribute key="nchroms" value="15" />
    ```
    """
    _assert_h5py()
    output_temp = io.BytesIO(output_bytes)
    with h5py.File(output_temp, "r", locking=False) as h5:
        local_attrs = h5.attrs
        assert (
            key in local_attrs and str(local_attrs[key]) == value
        ), f"Not a HDF5 file or H5 attributes do not match:\n\t{[(key, str(value)) for key, value in local_attrs.items()]}\n\n\t({key} : {value})"


# TODO the function actually queries groups. so the function and argument name are misleading
def assert_has_h5_keys(output_bytes: OutputBytes, keys: Keys) -> None:
    """Asserts the specified HDF5 output has the given keys."""
    _assert_h5py()
    h5_keys = sorted([k.strip() for k in keys.strip().split(",")])
    output_temp = io.BytesIO(output_bytes)
    local_keys = []

    def append_keys(key):
        local_keys.append(key)
        return None

    with h5py.File(output_temp, "r", locking=False) as f:
        f.visit(append_keys)
        missing = 0
        for key in h5_keys:
            if key not in local_keys:
                missing += 1
        assert missing == 0, f"Not a HDF5 file or H5 keys missing:\n\t{local_keys}\n\t{h5_keys}"



"""This module contains a linter for tool XML block order.

For more information on the IUC standard for XML block order see -
https://github.com/galaxy-iuc/standards.
"""

from typing import (
    Optional,
    TYPE_CHECKING,
)

from galaxy.tool_util.lint import Linter
from ._util import is_datasource

if TYPE_CHECKING:
    from galaxy.tool_util.lint import LintContext
    from galaxy.tool_util.parser.interface import ToolSource

# https://github.com/galaxy-iuc/standards
# https://github.com/galaxy-iuc/standards/pull/7/files
TAG_ORDER = [
    "description",
    "macros",
    "options",
    "edam_topics",
    "edam_operations",
    "xrefs",
    "parallelism",
    "requirements",
    "required_files",
    "code",
    "stdio",
    "version_command",
    "command",
    "environment_variables",
    "configfiles",
    "inputs",
    "outputs",
    "tests",
    "help",
    "citations",
]

DATASOURCE_TAG_ORDER = [
    "description",
    "macros",
    "requirements",
    "command",
    "configfiles",
    "inputs",
    "request_param_translation",
    "uihints",
    "outputs",
    "options",
    "help",
    "citations",
]


class XMLOrder(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        tool_root = tool_xml.getroot()

        if is_datasource(tool_xml):
            tag_ordering = DATASOURCE_TAG_ORDER
        else:
            tag_ordering = TAG_ORDER
        last_tag = None
        last_key: Optional[int] = None
        for elem in tool_root:
            tag = elem.tag
            if tag not in tag_ordering:
                continue
            key = tag_ordering.index(tag)
            if last_key:
                if last_key > key:
                    lint_ctx.warn(
                        f"Best practice violation [{tag}] elements should come before [{last_tag}]",
                        linter=cls.name(),
                        node=elem,
                    )
            last_tag = tag
            last_key = key



from typing import (
    Any,
    cast,
    Dict,
    Iterable,
    List,
    Optional,
    TypeVar,
    Union,
)

from typing_extensions import Protocol

from galaxy.tool_util_models.parameters import (
    ConditionalParameterModel,
    ConditionalWhen,
    simple_input_models,
    ToolParameterBundle,
    ToolParameterT,
)
from .state import ToolState


class VisitorNoReplacement:
    pass


class VisitorUndefined:
    pass


VISITOR_NO_REPLACEMENT = VisitorNoReplacement()
VISITOR_UNDEFINED = VisitorUndefined()


class Callback(Protocol):
    def __call__(self, parameter: ToolParameterT, value: Any):
        pass


def visit_input_values(
    input_models: ToolParameterBundle,
    tool_state: ToolState,
    callback: Callback,
    no_replacement_value=VISITOR_NO_REPLACEMENT,
) -> Dict[str, Any]:
    return _visit_input_values(
        simple_input_models(input_models.parameters),
        tool_state.input_state,
        callback=callback,
        no_replacement_value=no_replacement_value,
    )


def _visit_input_values(
    input_models: Iterable[ToolParameterT],
    input_values: Dict[str, Any],
    callback: Callback,
    no_replacement_value=VISITOR_NO_REPLACEMENT,
) -> Dict[str, Any]:

    def _callback(name: str, old_values: Dict[str, Any], new_values: Dict[str, Any]):
        input_value = old_values.get(name, VISITOR_UNDEFINED)
        if input_value is VISITOR_UNDEFINED:
            return
        replacement = callback(model, input_value)
        if replacement != no_replacement_value:
            new_values[name] = replacement
        else:
            new_values[name] = input_value

    new_input_values: Dict[str, Any] = {}
    for model in input_models:
        name = model.name
        input_value = input_values.get(name, VISITOR_UNDEFINED)
        if input_value is VISITOR_UNDEFINED:
            continue

        if model.parameter_type == "gx_repeat":
            repeat_parameters = model.parameters
            repeat_values = cast(list, input_value)
            new_repeat_values = []
            for repeat_instance_values in repeat_values:
                new_repeat_values.append(
                    _visit_input_values(
                        repeat_parameters, repeat_instance_values, callback, no_replacement_value=no_replacement_value
                    )
                )
            new_input_values[name] = new_repeat_values
        elif model.parameter_type == "gx_section":
            section_parameters = model.parameters
            section_values = cast(dict, input_value)
            new_section_values = _visit_input_values(
                section_parameters, section_values, callback, no_replacement_value=no_replacement_value
            )
            new_input_values[name] = new_section_values
        elif model.parameter_type == "gx_conditional":
            test_parameter = model.test_parameter
            test_parameter_name = test_parameter.name

            conditional_values = cast(dict, input_value)
            when: ConditionalWhen = _select_which_when(model, conditional_values)
            new_conditional_values = _visit_input_values(
                when.parameters, conditional_values, callback, no_replacement_value=no_replacement_value
            )
            if test_parameter_name in conditional_values:
                _callback(test_parameter_name, conditional_values, new_conditional_values)
            new_input_values[name] = new_conditional_values
        else:
            _callback(name, input_values, new_input_values)
    return new_input_values


def _select_which_when(conditional: ConditionalParameterModel, state: dict) -> ConditionalWhen:
    test_parameter = conditional.test_parameter
    test_parameter_name = test_parameter.name
    explicit_test_value = state.get(test_parameter_name)
    test_value = validate_explicit_conditional_test_value(test_parameter_name, explicit_test_value)
    for when in conditional.whens:
        print(when.discriminator)
        print(type(when.discriminator))
        if test_value is None and when.is_default_when:
            return when
        elif test_value == when.discriminator:
            return when
    else:
        raise Exception(f"Invalid conditional test value ({explicit_test_value}) for parameter ({test_parameter_name})")


def flat_state_path(has_name: Union[str, ToolParameterT], prefix: Optional[str] = None) -> str:
    """Given a parameter name or model and an optional prefix, give 'flat' name for parameter in tree."""
    if hasattr(has_name, "name"):
        name = cast(ToolParameterT, has_name).name
    else:
        name = has_name
    return name if prefix is None else f"{prefix}|{name}"


KVT = TypeVar("KVT")


def keys_starting_with(flat_tree: Dict[str, KVT], flat_state_path: str) -> Dict[str, KVT]:
    subset: Dict[str, KVT] = {}
    for key, value in flat_tree.items():
        if key.startswith(flat_state_path):
            subset[key] = value
    return subset


def repeat_inputs_to_array(flat_state_path: str, inputs: Dict[str, KVT]) -> List[Dict[str, KVT]]:
    repeat_inputs = keys_starting_with(inputs, flat_state_path + "_")
    highest_count = -1
    for key in repeat_inputs.keys():
        repeat_num_str = key[len(flat_state_path) + 1 :].split("|")[0]
        try:
            repeat_num = int(repeat_num_str)
            if repeat_num > highest_count:
                highest_count = repeat_num
        except ValueError:
            continue

    params: List[Dict[str, KVT]] = []
    for _ in range(highest_count + 1):
        instance_params: Dict[str, KVT] = {}
        params.append(instance_params)
    for key, value in repeat_inputs.items():
        repeat_num_str = key[len(flat_state_path) + 1 :].split("|")[0]
        try:
            repeat_num = int(repeat_num_str)
            params[repeat_num][key] = value
        except ValueError:
            continue
    return params


def validate_explicit_conditional_test_value(test_parameter_name: str, value: Any) -> Optional[Union[str, bool]]:
    if value is not None and not isinstance(value, (str, bool)):
        raise Exception(f"Invalid conditional test value ({value}) for parameter ({test_parameter_name})")
    return value



import io
import re
import tarfile
import tempfile
import zipfile

from galaxy.util import asbool
from ._types import (
    Annotated,
    AssertionParameter,
    ChildAssertions,
    Delta,
    Max,
    Min,
    N,
    Negate,
    NEGATE_DEFAULT,
    OutputBytes,
    VerifyAssertionsFunction,
    XmlBool,
)
from ._util import _assert_presence_number


def _extract_from_tar(output_bytes, fn):
    with io.BytesIO(output_bytes) as temp:
        with tarfile.open(fileobj=temp, mode="r") as tar_temp:
            ti = tar_temp.getmember(fn)
            # zip treats directories like empty files.
            # so make this consistent for tar
            if ti.isdir():
                return ""
            tar_file = tar_temp.extractfile(fn)
            assert tar_file is not None
            with tar_file as member_fh:
                return member_fh.read()


def _list_from_tar(output_bytes, path):
    lst = []
    with io.BytesIO(output_bytes) as temp:
        with tarfile.open(fileobj=temp, mode="r") as tar_temp:
            for fn in tar_temp.getnames():
                if not re.match(path, fn):
                    continue
                lst.append(fn)
    return sorted(lst)


def _extract_from_zip(output_bytes, fn):
    with io.BytesIO(output_bytes) as temp:
        with zipfile.ZipFile(temp, mode="r") as zip_temp:
            with zip_temp.open(fn) as member_fh:
                return member_fh.read()


def _list_from_zip(output_bytes, path):
    lst = []
    with io.BytesIO(output_bytes) as temp:
        with zipfile.ZipFile(temp, mode="r") as zip_temp:
            for fn in zip_temp.namelist():
                if not re.match(path, fn):
                    continue
                lst.append(fn)
    return sorted(lst)


Path = Annotated[str, AssertionParameter("The regular expression specifying the archive member.")]
All = Annotated[
    XmlBool,
    AssertionParameter(
        "Check the sub-assertions for all paths matching the path. Default: false, i.e. only the first",
        xml_type="PermissiveBoolean",
    ),
]


def assert_has_archive_member(
    output_bytes: OutputBytes,
    path: Path,
    verify_assertions_function: VerifyAssertionsFunction,
    children: ChildAssertions = None,
    all: All = False,
    n: N = None,
    delta: Delta = 0,
    min: Min = None,
    max: Max = None,
    negate: Negate = NEGATE_DEFAULT,
) -> None:
    """This tag allows to check if ``path`` is contained in a compressed file.

    The path is a regular expression that is matched against the full paths of the objects in
    the compressed file (remember that "matching" means it is checked if a prefix of
    the full path of an archive member is described by the regular expression).
    Valid archive formats include ``.zip``, ``.tar``, and ``.tar.gz``. Note that
    depending on the archive creation method:

    - full paths of the members may be prefixed with ``./``
    - directories may be treated as empty files

    ```xml
    <has_archive_member path="./path/to/my-file.txt"/>
    ```

    With ``n`` and ``delta`` (or ``min`` and ``max``) assertions on the number of
    archive members matching ``path`` can be expressed. The following could be used,
    e.g., to assert an archive containing n&plusmn;1 elements out of which at least
    4 need to have a ``txt`` extension.

    ```xml
    <has_archive_member path=".*" n="10" delta="1"/>
    <has_archive_member path=".*\\.txt" min="4"/>
    ```

    In addition the tag can contain additional assertions as child elements about
    the first member in the archive matching the regular expression ``path``. For
    instance

    ```xml
    <has_archive_member path=".*/my-file.txt">
      <not_has_text text="EDK72998.1"/>
    </has_archive_member>
    ```

    If the ``all`` attribute is set to ``true`` then all archive members are subject
    to the assertions. Note that, archive members matching the ``path`` are sorted
    alphabetically.

    The ``negate`` attribute of the ``has_archive_member`` assertion only affects
    the asserts on the presence and number of matching archive members, but not any
    sub-assertions (which can offer the ``negate`` attribute on their own).  The
    check if the file is an archive at all, which is also done by the function, is
    not affected."""
    all = asbool(all)
    extract_foo = None
    # from python 3.9 is_tarfile supports file like objects then we do not need
    # the tempfile detour but can use io.BytesIO(output_bytes)
    with tempfile.NamedTemporaryFile() as tmp:
        tmp.write(output_bytes)
        tmp.flush()
        if zipfile.is_zipfile(tmp.name):
            extract_foo = _extract_from_zip
            list_foo = _list_from_zip
        elif tarfile.is_tarfile(tmp.name):
            extract_foo = _extract_from_tar
            list_foo = _list_from_tar
    assert extract_foo is not None, f"Expected path '{path}' to be an archive"

    # get list of matching file names in archive and check against n, delta,
    # min, max (slightly abusing the output and text as well as the function
    # parameters)
    fns = list_foo(output_bytes, path)
    _assert_presence_number(
        None,
        path,
        n,
        delta,
        min,
        max,
        negate,
        lambda o, t: len(fns) > 0,
        lambda o, t: len(fns),
        "{expected} path '{text}' in archive",
        "{expected} {n}+-{delta} matches for path '{text}' in archive",
        "{expected} that the number of matches for path '{text}' in archive is in [{min}:{max}]",
    )

    # check sub-assertions on members matching path
    for fn in fns:
        contents = extract_foo(output_bytes, fn)
        try:
            verify_assertions_function(contents, children)
        except AssertionError as e:
            raise AssertionError(f"Archive member '{path}': {str(e)}")
        if not all:
            break



"""Linter for CWL tools."""

from typing import TYPE_CHECKING

lint_tool_types = ["cwl"]

from galaxy.tool_util.cwl.schema import schema_loader
from galaxy.tool_util.lint import Linter

if TYPE_CHECKING:
    from galaxy.tool_util.lint import LintContext
    from galaxy.tool_util.parser import ToolSource


class CWLValid(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        raw_reference = schema_loader.raw_process_reference(tool_source.source_path)
        try:
            schema_loader.process_definition(raw_reference)
        except Exception:
            return
        lint_ctx.info("CWL appears to be valid.", linter=cls.name())


class CWLInValid(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        raw_reference = schema_loader.raw_process_reference(tool_source.source_path)
        try:
            schema_loader.process_definition(raw_reference)
        except Exception as e:
            lint_ctx.error(f"Failed to valdiate CWL artifact: {e}", linter=cls.name())


class CWLVersionMissing(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        raw_reference = schema_loader.raw_process_reference(tool_source.source_path)
        cwl_version = raw_reference.process_object.get("cwlVersion", None)
        if cwl_version is None:
            lint_ctx.error("CWL file does not contain a 'cwlVersion'", linter=cls.name())


class CWLVersionUnknown(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        raw_reference = schema_loader.raw_process_reference(tool_source.source_path)
        cwl_version = raw_reference.process_object.get("cwlVersion", None)
        if cwl_version not in ["v1.0"]:
            lint_ctx.warn(
                f"CWL version [{cwl_version}] is unknown, we recommend the v1.0 the stable release.", linter=cls.name()
            )


class CWLVersionGood(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        raw_reference = schema_loader.raw_process_reference(tool_source.source_path)
        cwl_version = raw_reference.process_object.get("cwlVersion", None)
        if cwl_version in ["v1.0"]:
            lint_ctx.info(f"Modern CWL version [{cwl_version}].", linter=cls.name())


class CWLDockerMissing(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        _, containers, *_ = tool_source.parse_requirements()
        if len(containers) == 0:
            lint_ctx.warn("Tool does not specify a DockerPull source.", linter=cls.name())


class CWLDockerGood(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        _, containers, *_ = tool_source.parse_requirements()
        if len(containers) > 0:
            identifier = containers[0].identifier
            lint_ctx.info(f"Tool will run in Docker image [{identifier}].", linter=cls.name())


class CWLDescriptionMissing(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        help = tool_source.parse_help()
        if not help:
            lint_ctx.warn("Description of tool is empty or absent.", linter=cls.name())


class CWLHelpTODO(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        help = tool_source.parse_help()
        if help and "TODO" in help.content:
            lint_ctx.warn("Help contains TODO text.", linter=cls.name())



import os

from galaxy.tool_util.parameters import (
    input_models_for_tool_source,
    ToolParameterBundle,
    ToolParameterBundleModel,
    ToolParameterT,
)
from galaxy.tool_util.parser import (
    get_tool_source,
    ToolSource,
)
from galaxy.util import galaxy_directory
from . import functional_test_tool_path


class ParameterBundle(ToolParameterBundle):

    def __init__(self, parameter: ToolParameterT):
        self.parameters = [parameter]


def parameter_bundle(parameter: ToolParameterT) -> ParameterBundle:
    return ParameterBundle(parameter)


def parameter_bundle_for_framework_tool(filename: str) -> ToolParameterBundleModel:
    path = functional_test_tool_path(filename)
    tool_source = get_tool_source(path, macro_paths=[])
    return input_models_for_tool_source(tool_source)


def parameter_bundle_for_file(filename: str) -> ToolParameterBundleModel:
    tool_source = parameter_tool_source(filename)
    return input_models_for_tool_source(tool_source)


def parameter_tool_source(basename: str) -> ToolSource:
    path_prefix = os.path.join(galaxy_directory(), "test/functional/tools/parameters", basename)
    if os.path.exists(f"{path_prefix}.xml"):
        path = f"{path_prefix}.xml"
    else:
        path = f"{path_prefix}.cwl"
    tool_source = get_tool_source(path, macro_paths=[])
    return tool_source



"""Abstraction for waiting on API conditions to become true."""

from galaxy.util.wait import (
    timeout_type,
    TimeoutAssertionError,
    wait_on,
)

__all__ = ("wait_on", "TimeoutAssertionError", "timeout_type")



#!/usr/bin/env python

import argparse
import sys

import yaml

from galaxy.tool_util_models import Tests

DESCRIPTION = """
A small utility to verify the Planemo test format.

This script doesn't use semantic information about tools or workflows so only
the structure of the file is checked and things like inputs matching up is not
included.
"""


def validate_test_file(test_file: str) -> None:
    with open(test_file) as f:
        json = yaml.safe_load(f)
    Tests.model_validate(json)


def arg_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument("test_file")
    return parser


def main(argv=None) -> None:
    if argv is None:
        argv = sys.argv[1:]

    args = arg_parser().parse_args(argv)
    validate_test_file(args.test_file)


if __name__ == "__main__":
    main()



"""This module contains a linting functions for tool inputs."""

import ast
import re
import warnings
from copy import deepcopy
from typing import (
    Iterator,
    Optional,
    Tuple,
    TYPE_CHECKING,
)

from packaging.version import Version

from galaxy.tool_util.lint import Linter
from galaxy.util import string_as_bool
from ._util import (
    is_datasource,
    is_valid_cheetah_placeholder,
)
from ..parser.util import _parse_name

if TYPE_CHECKING:
    from galaxy.tool_util.lint import LintContext
    from galaxy.tool_util.parser import ToolSource
    from galaxy.util.etree import (
        Element,
        ElementTree,
    )

lint_tool_types = ["*"]

ATTRIB_VALIDATOR_COMPATIBILITY = {
    "check": ["metadata"],
    "expression": ["substitute_value_in_message"],
    "table_name": [
        "dataset_metadata_in_data_table",
        "dataset_metadata_not_in_data_table",
        "value_in_data_table",
        "value_not_in_data_table",
    ],
    "filename": ["dataset_metadata_in_file"],
    "metadata_name": [
        "dataset_metadata_equal",
        "dataset_metadata_in_data_table",
        "dataset_metadata_not_in_data_table",
        "dataset_metadata_in_file",
        "dataset_metadata_in_range",
    ],
    "metadata_column": [
        "dataset_metadata_in_data_table",
        "dataset_metadata_not_in_data_table",
        "value_in_data_table",
        "value_not_in_data_table",
        "dataset_metadata_in_file",
    ],
    "line_startswith": ["dataset_metadata_in_file"],
    "min": ["in_range", "length", "dataset_metadata_in_range"],
    "max": ["in_range", "length", "dataset_metadata_in_range"],
    "exclude_min": ["in_range", "dataset_metadata_in_range"],
    "exclude_max": ["in_range", "dataset_metadata_in_range"],
    "split": ["dataset_metadata_in_file"],
    "skip": ["metadata"],
    "value": ["dataset_metadata_equal"],
    "value_json": ["dataset_metadata_equal"],
}

PARAMETER_VALIDATOR_TYPE_COMPATIBILITY = {
    "integer": ["in_range", "expression"],
    "float": ["in_range", "expression"],
    "data": [
        "metadata",
        "no_options",
        "unspecified_build",
        "dataset_ok_validator",
        "dataset_metadata_equal",
        "dataset_metadata_in_range",
        "dataset_metadata_in_file",
        "dataset_metadata_in_data_table",
        "dataset_metadata_not_in_data_table",
        "expression",
    ],
    "data_collection": [
        "metadata",
        "no_options",
        "unspecified_build",
        "dataset_ok_validator",
        "dataset_metadata_equal",
        "dataset_metadata_in_range",
        "dataset_metadata_in_file",
        "dataset_metadata_in_data_table",
        "dataset_metadata_not_in_data_table",
        "expression",
    ],
    "text": ["regex", "length", "empty_field", "value_in_data_table", "value_not_in_data_table", "expression"],
    "select": [
        "in_range",
        "no_options",
        "regex",
        "length",
        "empty_field",
        "value_in_data_table",
        "value_not_in_data_table",
        "expression",
    ],
    "drill_down": [
        "no_options",
        "regex",
        "length",
        "empty_field",
        "value_in_data_table",
        "value_not_in_data_table",
        "expression",
    ],
    "data_column": [
        "no_options",
        "regex",
        "length",
        "empty_field",
        "value_in_data_table",
        "value_not_in_data_table",
        "expression",
    ],
}

PARAM_TYPE_CHILD_COMBINATIONS = [
    ("./options", ["data", "select", "drill_down"]),
    ("./options/option", ["drill_down"]),
    ("./options/column", ["data", "select"]),
]

# TODO lint for valid param type - attribute combinations
# TODO check if ref input param is present for from_dataset


class InputsNum(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        tool_node = tool_xml.find("./inputs")
        if tool_node is None:
            tool_node = tool_xml.getroot()
        num_inputs = len(tool_xml.findall("./inputs//param"))
        if num_inputs:
            lint_ctx.info(f"Found {num_inputs} input parameters.", linter=cls.name(), node=tool_node)


class InputsMissing(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        tool_node = tool_xml.find("./inputs")
        if tool_node is None:
            tool_node = tool_xml.getroot()
        num_inputs = len(tool_xml.findall("./inputs//param"))
        if num_inputs == 0 and not is_datasource(tool_xml):
            lint_ctx.warn("Found no input parameters.", linter=cls.name(), node=tool_node)


class InputsMissingDataSource(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        tool_node = tool_xml.find("./inputs")
        if tool_node is None:
            tool_node = tool_xml.getroot()
        num_inputs = len(tool_xml.findall("./inputs//param"))
        if num_inputs == 0 and is_datasource(tool_xml):
            lint_ctx.info("No input parameters, OK for data sources", linter=cls.name(), node=tool_node)


class InputsDatasourceTags(Linter):
    """
    Lint that datasource tools have display and uihints tags
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        tool_node = tool_xml.find("./inputs")
        if tool_node is None:
            tool_node = tool_xml.getroot()
        inputs = tool_xml.findall("./inputs//param")
        if is_datasource(tool_xml):
            # TODO only display is subtag of inputs, uihints is a separate top level tag (supporting only attrib minwidth)
            for datasource_tag in ("display", "uihints"):
                if not any(param.tag == datasource_tag for param in inputs):
                    lint_ctx.info(
                        f"{datasource_tag} tag usually present in data sources", linter=cls.name(), node=tool_node
                    )


def _iter_param(tool_xml: "ElementTree") -> Iterator[Tuple["Element", str]]:
    for param in tool_xml.findall("./inputs//param"):
        if "name" not in param.attrib and "argument" not in param.attrib:
            continue
        param_name = _parse_name(param.attrib.get("name"), param.attrib.get("argument"))
        yield param, param_name


def _iter_param_type(tool_xml: "ElementTree") -> Iterator[Tuple["Element", str, str]]:
    for param, param_name in _iter_param(tool_xml):
        if "type" not in param.attrib:
            continue
        param_type = param.attrib["type"]
        if param_type == "":
            continue
        yield param, param_name, param_type


class InputsName(Linter):
    """
    Lint params define a name
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param in tool_xml.findall("./inputs//param"):
            if "name" not in param.attrib and "argument" not in param.attrib:
                lint_ctx.error("Found param input with no name specified.", linter=cls.name(), node=param)


class InputsNameRedundantArgument(Linter):
    """
    Lint params define a name
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param in tool_xml.findall("./inputs//param"):
            if "name" in param.attrib and "argument" in param.attrib:
                param_name = _parse_name(param.attrib.get("name"), param.attrib.get("argument"))
                if param.attrib.get("name") == _parse_name(None, param.attrib.get("argument")):
                    lint_ctx.warn(
                        f"Param input [{param_name}] 'name' attribute is redundant if argument implies the same name.",
                        linter=cls.name(),
                        node=param,
                    )


# TODO redundant with InputsNameValid
class InputsNameEmpty(Linter):
    """
    Lint params define a non-empty name
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param, param_name in _iter_param(tool_xml):
            if param_name.strip() == "":
                lint_ctx.error("Param input with empty name.", linter=cls.name(), node=param)


class InputsNameValid(Linter):
    """
    Lint params define valid cheetah placeholder
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param, param_name in _iter_param(tool_xml):
            if param_name != "" and not is_valid_cheetah_placeholder(param_name):
                lint_ctx.warn(
                    f"Param input [{param_name}] is not a valid Cheetah placeholder.", linter=cls.name(), node=param
                )


def _param_path(param: "Element", param_name: str) -> str:
    path = [param_name]
    parent = param
    while True:
        parent = parent.getparent()
        if parent.tag == "inputs":
            break
        # parameters of the same name in different when branches are allowed
        # just add the value of the when branch to the path (this also allows
        # that the conditional select has the same name as params in the whens)
        if parent.tag == "when":
            path.append(str(parent.attrib.get("value")))
        else:
            path.append(str(parent.attrib.get("name")))
    return ".".join(reversed(path))


class InputsNameDuplicate(Linter):
    """
    Lint params with duplicate names
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        input_names = set()
        for param, param_name in _iter_param(tool_xml):
            # check for parameters with duplicate names
            path = _param_path(param, param_name)
            if path in input_names:
                lint_ctx.error(
                    f"Tool defines multiple parameters with the same name: '{path}'", linter=cls.name(), node=param
                )
            input_names.add(path)


class InputsNameDuplicateOutput(Linter):
    """
    Lint params names that are also used in outputs
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        input_names = set()
        for param, param_name in _iter_param(tool_xml):
            input_names.add(_param_path(param, param_name))

        # check if there is an output with the same name as an input
        outputs = tool_xml.findall("./outputs/*")
        for output in outputs:
            if output.get("name") in input_names:
                lint_ctx.error(
                    f'Tool defines an output with a name equal to the name of an input: \'{output.get("name")}\'',
                    linter=cls.name(),
                    node=output,
                )


class InputsTypeChildCombination(Linter):
    """
    Lint for invalid parameter type child combinations
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param, param_name, param_type in _iter_param_type(tool_xml):
            # lint for valid param type - child node combinations
            for ptcc in PARAM_TYPE_CHILD_COMBINATIONS:
                if param.find(ptcc[0]) is not None and param_type not in ptcc[1]:
                    lint_ctx.error(
                        f"Parameter [{param_name}] '{ptcc[0]}' tags are only allowed for parameters of type {ptcc[1]}",
                        linter=cls.name(),
                        node=param,
                    )


class InputsDataFormat(Linter):
    """
    Lint for data params wo format
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param, param_name, param_type in _iter_param_type(tool_xml):
            if param_type != "data":
                continue
            if "format" not in param.attrib:
                lint_ctx.warn(
                    f"Param input [{param_name}] with no format specified - 'data' format will be assumed.",
                    linter=cls.name(),
                    node=param,
                )


class InputsDataOptionsMultiple(Linter):
    """
    Lint for data params with multiple options tags
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param, param_name, param_type in _iter_param_type(tool_xml):
            if param_type != "data":
                continue
            options = param.findall("./options")
            if len(options) > 1:
                lint_ctx.error(
                    f"Data parameter [{param_name}] contains multiple options elements.",
                    linter=cls.name(),
                    node=options[1],
                )


class InputsDataOptionsAttrib(Linter):
    """
    Lint for data params with options that have invalid attributes
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param, param_name, param_type in _iter_param_type(tool_xml):
            if param_type != "data":
                continue
            options = param.find("./options")
            if options is None:
                continue
            for oa in options.attrib:
                if oa != "options_filter_attribute":
                    lint_ctx.error(
                        f"Data parameter [{param_name}] uses invalid attribute: {oa}", linter=cls.name(), node=param
                    )


class InputsDataOptionsFilterAttribFiltersType(Linter):
    """
    Lint for valid filter types for data parameters
    if options set options_filter_attribute
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param, param_name, param_type in _iter_param_type(tool_xml):
            if param_type != "data":
                continue
            options = param.find("./options")
            if options is None:
                continue
            for filter in param.findall("./options/filter"):
                if "options_filter_attribute" in options.attrib:
                    if filter.get("type") != "data_meta":
                        lint_ctx.error(
                            f'Data parameter [{param_name}] for filters only type="data_meta" is allowed, found type="{filter.get("type")}"',
                            linter=cls.name(),
                            node=filter,
                        )


class InputsDataOptionsFiltersType(Linter):
    """
    Lint for valid filter types for data parameters
    if options do not set options_filter_attribute
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param, param_name, param_type in _iter_param_type(tool_xml):
            if param_type != "data":
                continue
            options = param.find("./options")
            if options is None:
                continue
            for filter in param.findall("./options/filter"):
                if "options_filter_attribute" not in options.attrib:
                    if filter.get("key") != "dbkey" or filter.get("type") != "data_meta":
                        lint_ctx.error(
                            f'Data parameter [{param_name}] for filters only type="data_meta" and key="dbkey" are allowed, found type="{filter.get("type")}" and key="{filter.get("key")}"',
                            linter=cls.name(),
                            node=filter,
                        )


FILTER_REQUIRED_ATTRIBUTES = {
    "data_meta": ["type", "ref", "key"],  # column needs special treatment
    "param_value": ["type", "ref", "column"],
    "static_value": ["type", "column", "value"],
    "regexp": ["type", "column", "value"],
    "unique_value": ["type", "column"],
    "multiple_splitter": ["type", "column"],
    "attribute_value_splitter": ["type", "column"],
    "add_value": ["type", "value"],
    "remove_value": ["type"],  # this is handled separately in InputsOptionsRemoveValueFilterRequiredAttributes
    "sort_by": ["type", "column"],
    "data_table": ["type", "column", "table_name", "data_table_column"],
}


class InputsOptionsFiltersRequiredAttributes(Linter):
    """
    check required attributes of filters
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param, param_name in _iter_param(tool_xml):
            options = param.find("./options")
            if options is None:
                continue
            for filter in param.findall("./options/filter"):
                filter_type = filter.get("type", None)
                if filter_type is None or filter_type not in FILTER_ALLOWED_ATTRIBUTES:
                    continue
                for attrib in FILTER_REQUIRED_ATTRIBUTES[filter_type]:
                    if attrib not in filter.attrib:
                        lint_ctx.error(
                            f"Select parameter [{param_name}] '{filter_type}' filter misses required attribute '{attrib}'",
                            linter=cls.name(),
                            node=filter,
                        )


class InputsOptionsRemoveValueFilterRequiredAttributes(Linter):
    """
    check required attributes of remove_value filter
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param, param_name in _iter_param(tool_xml):
            options = param.find("./options")
            if options is None:
                continue
            for filter in param.findall("./options/filter"):
                filter_type = filter.get("type", None)
                # check for required attributes for filter (remove_value needs a bit more logic here)
                if filter_type != "remove_value":
                    continue
                if not (
                    (
                        "value" in filter.attrib
                        and "ref" not in filter.attrib
                        and "meta_ref" not in filter.attrib
                        and "key" not in filter.attrib
                    )
                    or (
                        "value" not in filter.attrib
                        and "ref" in filter.attrib
                        and "meta_ref" not in filter.attrib
                        and "key" not in filter.attrib
                    )
                    or (
                        "value" not in filter.attrib
                        and "ref" not in filter.attrib
                        and "meta_ref" in filter.attrib
                        and "key" in filter.attrib
                    )
                ):
                    lint_ctx.error(
                        f"Select parameter [{param_name}] '{filter_type}'' filter needs either the 'value'; 'ref'; or 'meta' and 'key' attribute(s)",
                        linter=cls.name(),
                        node=filter,
                    )


FILTER_ALLOWED_ATTRIBUTES = deepcopy(FILTER_REQUIRED_ATTRIBUTES)
FILTER_ALLOWED_ATTRIBUTES["static_value"].append("keep")
FILTER_ALLOWED_ATTRIBUTES["regexp"].append("keep")
FILTER_ALLOWED_ATTRIBUTES["data_meta"].extend(["column", "multiple", "separator"])
FILTER_ALLOWED_ATTRIBUTES["param_value"].extend(["keep", "ref_attribute"])
FILTER_ALLOWED_ATTRIBUTES["multiple_splitter"].append("separator")
FILTER_ALLOWED_ATTRIBUTES["attribute_value_splitter"].extend(["pair_separator", "name_val_separator"])
FILTER_ALLOWED_ATTRIBUTES["add_value"].extend(["name", "index"])
FILTER_ALLOWED_ATTRIBUTES["remove_value"].extend(["value", "ref", "meta_ref", "key"])
FILTER_ALLOWED_ATTRIBUTES["data_table"].append("keep")
FILTER_ALLOWED_ATTRIBUTES["sort_by"].append("reverse_sort_order")


class InputsOptionsFiltersAllowedAttributes(Linter):
    """
    check allowed attributes of filters
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return

        for param, param_name in _iter_param(tool_xml):
            options = param.find("./options")
            if options is None:
                continue
            for filter in param.findall("./options/filter"):
                filter_type = filter.get("type", None)
                if filter_type is None or filter_type not in FILTER_ALLOWED_ATTRIBUTES:
                    continue
                for attrib in filter.attrib:
                    if attrib not in FILTER_ALLOWED_ATTRIBUTES[filter_type]:
                        lint_ctx.warn(
                            f"Select parameter [{param_name}] '{filter_type}' filter specifies unnecessary attribute '{attrib}'",
                            linter=cls.name(),
                            node=filter,
                        )


class InputsOptionsRegexFilterExpression(Linter):
    """
    Check the regular expression of regexp filters
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return

        for param, param_name in _iter_param(tool_xml):
            options = param.find("./options")
            if options is None:
                continue
            for filter in param.findall("./options/filter"):
                filter_type = filter.get("type", None)
                if filter_type == "regexp" and "value" in filter.attrib:
                    try:
                        re.compile(filter.attrib["value"])
                    except re.error as re_error:
                        lint_ctx.error(
                            f"Select parameter [{param_name}] '{filter_type}'' filter 'value' is not a valid regular expression ({re_error})'",
                            linter=cls.name(),
                            node=filter,
                        )


class InputsOptionsFiltersCheckReferences(Linter):
    """
    Check the references used in filters
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return

        # get the set of param names
        param_names = {param_name for _, param_name in _iter_param(tool_xml)}

        for param, param_name in _iter_param(tool_xml):
            options = param.find("./options")
            if options is None:
                continue
            for filter in param.findall("./options/filter"):
                filter_type = filter.get("type", None)
                if filter_type is not None:
                    # check for references to other inputs
                    # TODO: currently ref and metaref seem only to work for top level params,
                    # once this is fixed the linter needs to be extended, e.g. `f.attrib[ref_attrib].split('|')[-1]`
                    for ref_attrib in ["meta_ref", "ref"]:
                        if ref_attrib in filter.attrib and filter.attrib[ref_attrib] not in param_names:
                            lint_ctx.error(
                                f"Select parameter [{param_name}] '{filter_type}'' filter attribute '{ref_attrib}' refers to non existing parameter '{filter.attrib[ref_attrib]}'",
                                linter=cls.name(),
                                node=filter,
                            )


class InputsDataOptionsFiltersRef(Linter):
    """
    Lint for set ref for filters of data parameters
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param, param_name, param_type in _iter_param_type(tool_xml):
            if param_type != "data":
                continue
            options = param.find("./options")
            if options is None:
                continue
            for filter in param.findall("./options/filter"):
                if not filter.get("ref"):
                    lint_ctx.error(
                        f"Data parameter [{param_name}] filter needs to define a ref attribute",
                        linter=cls.name(),
                        node=filter,
                    )


class InputsSelectDynamicOptions(Linter):
    """
    Lint for select with deprecated dynamic_options attribute
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param, param_name, param_type in _iter_param_type(tool_xml):
            if param_type != "select":
                continue
            dynamic_options = param.get("dynamic_options", None)
            if dynamic_options is not None:
                lint_ctx.warn(
                    f"Select parameter [{param_name}] uses deprecated 'dynamic_options' attribute.",
                    linter=cls.name(),
                    node=param,
                )


class InputsSelectOptionsDef(Linter):
    """
    Lint for valid ways to define select options
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param, param_name, param_type in _iter_param_type(tool_xml):
            if param_type != "select":
                continue
            dynamic_options = param.get("dynamic_options", None)
            options = param.findall("./options")
            select_options = param.findall("./option")
            if param.getparent().tag != "conditional":
                if (dynamic_options is not None) + (len(options) > 0) + (len(select_options) > 0) != 1:
                    lint_ctx.error(
                        f"Select parameter [{param_name}] options have to be defined by either 'option' children elements, a 'options' element or the 'dynamic_options' attribute.",
                        linter=cls.name(),
                        node=param,
                    )


class InputsSelectOptionsDefConditional(Linter):
    """
    Lint for valid ways to define select options
    (for a select in a conditional)
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param, param_name, param_type in _iter_param_type(tool_xml):
            if param_type != "select":
                continue
            dynamic_options = param.get("dynamic_options", None)
            options = param.findall("./options")
            select_options = param.findall("./option")
            if param.getparent().tag == "conditional":
                if len(select_options) == 0 or dynamic_options is not None or len(options) > 0:
                    lint_ctx.error(
                        f"Select parameter of a conditional [{param_name}] options have to be defined by 'option' children elements.",
                        linter=cls.name(),
                        node=param,
                    )


# TODO xsd
class InputsSelectOptionValueMissing(Linter):
    """
    Lint for select option tags without value
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param, param_name, param_type in _iter_param_type(tool_xml):
            if param_type != "select":
                continue
            select_options = param.findall("./option")
            if any("value" not in option.attrib for option in select_options):
                lint_ctx.error(
                    f"Select parameter [{param_name}] has option without value", linter=cls.name(), node=param
                )


class InputsSelectOptionDuplicateValue(Linter):
    """
    Lint for select option with same value
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param, param_name, param_type in _iter_param_type(tool_xml):
            if param_type != "select":
                continue
            select_options = param.findall("./option")
            select_options_values = []
            for option in select_options:
                value = option.attrib.get("value", "")
                select_options_values.append((value, option.attrib.get("selected", "false")))
            if len(set(select_options_values)) != len(select_options_values):
                lint_ctx.error(
                    f"Select parameter [{param_name}] has multiple options with the same value",
                    linter=cls.name(),
                    node=param,
                )


class InputsSelectOptionDuplicateText(Linter):
    """
    Lint for select option with same text
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param, param_name, param_type in _iter_param_type(tool_xml):
            if param_type != "select":
                continue
            select_options = param.findall("./option")
            select_options_texts = []
            for option in select_options:
                if option.text is None:
                    text = option.attrib.get("value", "").capitalize()
                else:
                    text = option.text
                select_options_texts.append((text, option.attrib.get("selected", "false")))
            if len(set(select_options_texts)) != len(select_options_texts):
                lint_ctx.error(
                    f"Select parameter [{param_name}] has multiple options with the same text content",
                    linter=cls.name(),
                    node=param,
                )


class InputsSelectOptionsMultiple(Linter):
    """
    Lint dynamic options select for multiple options tags
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param, param_name, param_type in _iter_param_type(tool_xml):
            if param_type != "select":
                continue
            options = param.findall("./options")
            if len(options) > 1:
                lint_ctx.error(
                    f"Select parameter [{param_name}] contains multiple options elements.",
                    linter=cls.name(),
                    node=options[1],
                )


class InputsSelectOptionsDefinesOptions(Linter):
    """
    Lint dynamic options select for the potential to define options
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param, param_name, param_type in _iter_param_type(tool_xml):
            if param_type != "select":
                continue
            options = param.find("./options")
            if options is None:
                continue
            filter_adds_options = any(
                f.get("type", None) in ["add_value", "data_meta"] for f in param.findall("./options/filter")
            )
            from_file = options.get("from_file", None)
            from_parameter = options.get("from_parameter", None)
            # TODO check if input param is present for from_dataset
            from_dataset = options.get("from_dataset", None)
            from_data_table = options.get("from_data_table", None)
            from_url = options.get("from_url", None)

            if (
                from_file is None
                and from_parameter is None
                and from_dataset is None
                and from_data_table is None
                and from_url is None
                and not filter_adds_options
            ):
                lint_ctx.error(
                    f"Select parameter [{param_name}] options tag defines no options. Use 'from_dataset', 'from_data_table', or a filter that adds values.",
                    linter=cls.name(),
                    node=options,
                )


class InputsSelectOptionsDeprecatedAttr(Linter):
    """
    Lint dynamic options select deprecated attributes
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param, param_name, param_type in _iter_param_type(tool_xml):
            if param_type != "select":
                continue
            options = param.find("./options")
            if options is None:
                continue
            for deprecated_attr in ["from_file", "from_parameter", "options_filter_attribute", "transform_lines"]:
                if options.get(deprecated_attr) is not None:
                    lint_ctx.warn(
                        f"Select parameter [{param_name}] options uses deprecated '{deprecated_attr}' attribute.",
                        linter=cls.name(),
                        node=options,
                    )


class InputsSelectOptionsFromDatasetAndDatatable(Linter):
    """
    Lint dynamic options select for from_dataset and from_data_table
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param, param_name, param_type in _iter_param_type(tool_xml):
            if param_type != "select":
                continue
            options = param.find("./options")
            if options is None:
                continue
            from_dataset = options.get("from_dataset", None)
            from_data_table = options.get("from_data_table", None)
            if from_dataset is not None and from_data_table is not None:
                lint_ctx.error(
                    f"Select parameter [{param_name}] options uses 'from_dataset' and 'from_data_table' attribute.",
                    linter=cls.name(),
                    node=options,
                )


class InputsSelectOptionsMetaFileKey(Linter):
    """
    Lint dynamic options select: meta_file_key attribute can only be used with from_dataset
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param, param_name, param_type in _iter_param_type(tool_xml):
            if param_type != "select":
                continue
            options = param.find("./options")
            if options is None:
                continue
            from_dataset = options.get("from_dataset", None)
            if options.get("meta_file_key", None) is not None and from_dataset is None:
                lint_ctx.error(
                    f"Select parameter [{param_name}] 'meta_file_key' is only compatible with 'from_dataset'.",
                    linter=cls.name(),
                    node=options,
                )


class InputsBoolDistinctValues(Linter):
    """
    Lint booleans for distinct true/false value
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param, param_name, param_type in _iter_param_type(tool_xml):
            if param_type != "boolean":
                continue
            profile = tool_source.parse_profile()
            truevalue = param.attrib.get("truevalue", "true")
            falsevalue = param.attrib.get("falsevalue", "false")
            problematic_booleans_allowed = Version(profile) < Version("23.1")
            lint_level = lint_ctx.warn if problematic_booleans_allowed else lint_ctx.error
            if truevalue == falsevalue:
                lint_level(
                    f"Boolean parameter [{param_name}] needs distinct 'truevalue' and 'falsevalue' values.",
                    linter=cls.name(),
                    node=param,
                )


class InputsBoolProblematic(Linter):
    """
    Lint booleans for problematic values, i.e. truevalue being false and vice versa
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param, param_name, param_type in _iter_param_type(tool_xml):
            if param_type != "boolean":
                continue
            profile = tool_source.parse_profile()
            truevalue = param.attrib.get("truevalue", "true")
            falsevalue = param.attrib.get("falsevalue", "false")
            problematic_booleans_allowed = Version(profile) < Version("23.1")
            lint_level = lint_ctx.warn if problematic_booleans_allowed else lint_ctx.error
            if truevalue.lower() == "false":
                lint_level(
                    f"Boolean parameter [{param_name}] has invalid truevalue [{truevalue}].",
                    linter=cls.name(),
                    node=param,
                )
            if falsevalue.lower() == "true":
                lint_level(
                    f"Boolean parameter [{param_name}] has invalid falsevalue [{falsevalue}].",
                    linter=cls.name(),
                    node=param,
                )


class InputsSelectSingleCheckboxes(Linter):
    """
    Lint selects that allow only a single selection but display as checkboxes
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param, param_name, param_type in _iter_param_type(tool_xml):
            if param_type not in ["select", "data_column", "drill_down"]:
                continue
            multiple = string_as_bool(param.attrib.get("multiple", "false"))
            if param.attrib.get("display") == "checkboxes":
                if not multiple:
                    lint_ctx.error(
                        f'Select [{param_name}] `display="checkboxes"` is incompatible with `multiple="false"`, remove the `display` attribute',
                        linter=cls.name(),
                        node=param,
                    )


class InputsSelectMandatoryCheckboxes(Linter):
    """
    Lint selects that are mandatory but display as checkboxes
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param, param_name, param_type in _iter_param_type(tool_xml):
            if param_type not in ["select", "data_column", "drill_down"]:
                continue
            multiple = string_as_bool(param.attrib.get("multiple", "false"))
            optional = string_as_bool(param.attrib.get("optional", multiple))
            if param.attrib.get("display") == "checkboxes":
                if not optional:
                    lint_ctx.error(
                        f'Select [{param_name}] `display="checkboxes"` is incompatible with `optional="false"`, remove the `display` attribute',
                        linter=cls.name(),
                        node=param,
                    )


class InputsSelectMultipleRadio(Linter):
    """
    Lint selects that allow only multiple selections but display as radio
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param, param_name, param_type in _iter_param_type(tool_xml):
            if param_type not in ["select", "data_column", "drill_down"]:
                continue
            multiple = string_as_bool(param.attrib.get("multiple", "false"))
            if param.attrib.get("display") == "radio":
                if multiple:
                    lint_ctx.error(
                        f'Select [{param_name}] display="radio" is incompatible with multiple="true"',
                        linter=cls.name(),
                        node=param,
                    )


class InputsSelectOptionalRadio(Linter):
    """
    Lint selects that are optional but display as radio
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param, param_name, param_type in _iter_param_type(tool_xml):
            if param_type not in ["select", "data_column", "drill_down"]:
                continue
            multiple = string_as_bool(param.attrib.get("multiple", "false"))
            optional = string_as_bool(param.attrib.get("optional", multiple))
            if param.attrib.get("display") == "radio":
                if optional:
                    lint_ctx.error(
                        f'Select [{param_name}] display="radio" is incompatible with optional="true"',
                        linter=cls.name(),
                        node=param,
                    )


def _iter_param_validator(tool_xml: "ElementTree") -> Iterator[Tuple[str, str, "Element", str]]:
    input_params = tool_xml.findall("./inputs//param[@type]")
    for param in input_params:
        try:
            param_name = _parse_name(param.attrib.get("name"), param.attrib.get("argument"))
        except ValueError:
            continue
        param_type = param.attrib["type"]
        validators = param.findall("./validator[@type]")
        for validator in validators:
            vtype = validator.attrib["type"]
            yield (param_name, param_type, validator, vtype)


class ValidatorParamIncompatible(Linter):
    """
    Lint for validator type - parameter type incompatibilities
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param_name, param_type, validator, vtype in _iter_param_validator(tool_xml):
            if param_type in PARAMETER_VALIDATOR_TYPE_COMPATIBILITY:
                if vtype not in PARAMETER_VALIDATOR_TYPE_COMPATIBILITY[param_type]:
                    lint_ctx.error(
                        f"Parameter [{param_name}]: validator with an incompatible type '{vtype}'",
                        linter=cls.name(),
                        node=validator,
                    )


class ValidatorAttribIncompatible(Linter):
    """
    Lint for incompatibilities between validator type and given attributes
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param_name, _, validator, vtype in _iter_param_validator(tool_xml):
            for attrib in ATTRIB_VALIDATOR_COMPATIBILITY:
                if attrib in validator.attrib and vtype not in ATTRIB_VALIDATOR_COMPATIBILITY[attrib]:
                    lint_ctx.error(
                        f"Parameter [{param_name}]: attribute '{attrib}' is incompatible with validator of type '{vtype}'",
                        linter=cls.name(),
                        node=validator,
                    )


class ValidatorHasText(Linter):
    """
    Lint that parameters that need text have text
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param_name, _, validator, vtype in _iter_param_validator(tool_xml):
            if vtype in ["expression", "regex"]:
                if validator.text is None:
                    lint_ctx.error(
                        f"Parameter [{param_name}]: {vtype} validators are expected to contain text",
                        linter=cls.name(),
                        node=validator,
                    )


class ValidatorHasNoText(Linter):
    """
    Lint that parameters that need no text have no text
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param_name, _, validator, vtype in _iter_param_validator(tool_xml):
            if vtype not in ["expression", "regex"] and validator.text is not None:
                lint_ctx.warn(
                    f"Parameter [{param_name}]: '{vtype}' validators are not expected to contain text (found '{validator.text}')",
                    linter=cls.name(),
                    node=validator,
                )


class ValidatorExpression(Linter):
    """
    Lint that checks expressions / regexp (ignoring FutureWarning)
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        with warnings.catch_warnings():
            warnings.simplefilter("error", FutureWarning)
            for param_name, _, validator, vtype in _iter_param_validator(tool_xml):
                if vtype in ["expression", "regex"] and validator.text is not None:
                    try:
                        if vtype == "regex":
                            re.compile(validator.text)
                        else:
                            ast.parse(validator.text, mode="eval")
                    except FutureWarning:
                        pass
                    except Exception as e:
                        lint_ctx.error(
                            f"Parameter [{param_name}]: '{validator.text}' is no valid {vtype}: {str(e)}",
                            linter=cls.name(),
                            node=validator,
                        )


class ValidatorExpressionFuture(Linter):
    """
    Lint that checks expressions / regexp FutureWarnings
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        with warnings.catch_warnings():
            warnings.simplefilter("error", FutureWarning)
            for param_name, _, validator, vtype in _iter_param_validator(tool_xml):
                if vtype in ["expression", "regex"] and validator.text is not None:
                    try:
                        if vtype == "regex":
                            re.compile(validator.text)
                        else:
                            ast.parse(validator.text, mode="eval")
                    except FutureWarning as e:
                        lint_ctx.warn(
                            f"Parameter [{param_name}]: '{validator.text}' is marked as deprecated {vtype}: {str(e)}",
                            linter=cls.name(),
                            node=validator,
                        )
                    except Exception:
                        pass


class ValidatorMinMax(Linter):
    """
    Lint for min/max for validator that need it
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param_name, _, validator, vtype in _iter_param_validator(tool_xml):
            if vtype in ["in_range", "length", "dataset_metadata_in_range"] and (
                "min" not in validator.attrib and "max" not in validator.attrib
            ):
                lint_ctx.error(
                    f"Parameter [{param_name}]: '{vtype}' validators need to define the 'min' or 'max' attribute(s)",
                    linter=cls.name(),
                    node=validator,
                )


class ValidatorDatasetMetadataEqualValue(Linter):
    """
    Lint dataset_metadata_equal needs value or value_json
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param_name, _, validator, vtype in _iter_param_validator(tool_xml):
            if vtype in ["dataset_metadata_equal"]:
                if (
                    not ("value" in validator.attrib or "value_json" in validator.attrib)
                    or "metadata_name" not in validator.attrib
                ):
                    lint_ctx.error(
                        f"Parameter [{param_name}]: '{vtype}' validators need to define the 'value'/'value_json' and 'metadata_name' attributes",
                        linter=cls.name(),
                        node=validator,
                    )


class ValidatorDatasetMetadataEqualValueOrJson(Linter):
    """
    Lint dataset_metadata_equal needs either value or value_json
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param_name, _, validator, vtype in _iter_param_validator(tool_xml):
            if vtype in ["dataset_metadata_equal"]:
                if "value" in validator.attrib and "value_json" in validator.attrib:
                    lint_ctx.error(
                        f"Parameter [{param_name}]: '{vtype}' validators must not define the 'value' and the 'value_json' attributes",
                        linter=cls.name(),
                        node=validator,
                    )


class ValidatorMetadataCheckSkip(Linter):
    """
    Lint metadata needs check or skip
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param_name, _, validator, vtype in _iter_param_validator(tool_xml):
            if vtype in ["metadata"] and ("check" not in validator.attrib and "skip" not in validator.attrib):
                lint_ctx.error(
                    f"Parameter [{param_name}]: '{vtype}' validators need to define the 'check' or 'skip' attribute(s)",
                    linter=cls.name(),
                    node=validator,
                )


class ValidatorTableName(Linter):
    """
    Lint table_name is present if needed
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param_name, _, validator, vtype in _iter_param_validator(tool_xml):
            if (
                vtype
                in [
                    "value_in_data_table",
                    "value_not_in_data_table",
                    "dataset_metadata_in_data_table",
                    "dataset_metadata_not_in_data_table",
                ]
                and "table_name" not in validator.attrib
            ):
                lint_ctx.error(
                    f"Parameter [{param_name}]: '{vtype}' validators need to define the 'table_name' attribute",
                    linter=cls.name(),
                    node=validator,
                )


class ValidatorMetadataName(Linter):
    """
    Lint metadata_name is present if needed
    """

    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for param_name, _, validator, vtype in _iter_param_validator(tool_xml):
            if (
                vtype
                in [
                    "dataset_metadata_in_data_table",
                    "dataset_metadata_not_in_data_table",
                    "dataset_metadata_in_file",
                    "dataset_metadata_in_range",
                ]
                and "metadata_name" not in validator.attrib
            ):
                lint_ctx.error(
                    f"Parameter [{param_name}]: '{vtype}' validators need to define the 'metadata_name' attribute",
                    linter=cls.name(),
                    node=validator,
                )


def _iter_conditional(tool_xml: "ElementTree") -> Iterator[Tuple["Element", Optional[str], "Element", Optional[str]]]:
    conditionals = tool_xml.findall("./inputs//conditional")
    for conditional in conditionals:
        conditional_name = conditional.get("name")
        if conditional.get("value_from"):  # Probably only the upload tool use this, no children elements
            continue
        first_param = conditional.find("param")
        if first_param is None:
            continue
        first_param_type = first_param.get("type")
        yield conditional, conditional_name, first_param, first_param_type


class ConditionalParamTypeBool(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for _, conditional_name, first_param, first_param_type in _iter_conditional(tool_xml):
            if first_param_type == "boolean":
                lint_ctx.warn(
                    f'Conditional [{conditional_name}] first param of type="boolean" is discouraged, use a select',
                    linter=cls.name(),
                    node=first_param,
                )


class ConditionalParamType(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for _, conditional_name, first_param, first_param_type in _iter_conditional(tool_xml):
            if first_param_type not in ["boolean", "select"]:
                lint_ctx.error(
                    f'Conditional [{conditional_name}] first param should have type="select"',
                    linter=cls.name(),
                    node=first_param,
                )


class ConditionalParamIncompatibleAttributes(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for _, conditional_name, first_param, first_param_type in _iter_conditional(tool_xml):
            if first_param_type not in ["boolean", "select"]:
                continue
            for incomp in ["optional", "multiple"]:
                if string_as_bool(first_param.get(incomp, False)):
                    lint_ctx.warn(
                        f'Conditional [{conditional_name}] test parameter cannot be {incomp}="true"',
                        linter=cls.name(),
                        node=first_param,
                    )


class ConditionalWhenMissing(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for conditional, conditional_name, first_param, first_param_type in _iter_conditional(tool_xml):
            if first_param_type not in ["boolean", "select"]:
                continue
            if first_param_type == "select":
                options = first_param.findall("./option[@value]")
                option_ids = {option.get("value") for option in options}
            else:  # boolean
                option_ids = {first_param.get("truevalue", "true"), first_param.get("falsevalue", "false")}
            whens = conditional.findall("./when[@value]")
            when_ids = {w.get("value") for w in whens if w.get("value") is not None}
            for option_id in option_ids - when_ids:
                lint_ctx.warn(
                    f"Conditional [{conditional_name}] no <when /> block found for {first_param_type} option '{option_id}'",
                    linter=cls.name(),
                    node=conditional,
                )


class ConditionalOptionMissing(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for conditional, conditional_name, first_param, first_param_type in _iter_conditional(tool_xml):
            if first_param_type != "select":
                continue
            options = first_param.findall("./option[@value]")
            option_ids = {option.get("value") for option in options}
            whens = conditional.findall("./when[@value]")
            when_ids = {w.get("value") for w in whens if w.get("value") is not None}
            for when_id in when_ids - option_ids:
                lint_ctx.warn(
                    f"Conditional [{conditional_name}] no <option /> found for when block '{when_id}'",
                    linter=cls.name(),
                    node=conditional,
                )


class ConditionalOptionMissingBoolean(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        for conditional, conditional_name, first_param, first_param_type in _iter_conditional(tool_xml):
            if first_param_type != "boolean":
                continue
            option_ids = {first_param.get("truevalue", "true"), first_param.get("falsevalue", "false")}
            whens = conditional.findall("./when[@value]")
            when_ids = {w.get("value") for w in whens if w.get("value")}
            for when_id in when_ids - option_ids:
                lint_ctx.warn(
                    f"Conditional [{conditional_name}] no truevalue/falsevalue found for when block '{when_id}'",
                    linter=cls.name(),
                    node=conditional,
                )



from typing import (
    Any,
    cast,
    Dict,
    List,
    Optional,
    Union,
)

from galaxy.tool_util.parser.cwl import CwlInputSource
from galaxy.tool_util.parser.interface import (
    InputSource,
    PageSource,
    PagesSource,
    ToolSource,
)
from galaxy.tool_util.parser.parameter_validators import static_validators
from galaxy.tool_util.parser.util import (
    multiple_select_value_split,
    parse_profile_version,
    text_input_is_optional,
)
from galaxy.tool_util_models.parameters import (
    BaseUrlParameterModel,
    BooleanParameterModel,
    ColorParameterModel,
    cond_test_parameter_default_value,
    ConditionalParameterModel,
    ConditionalWhen,
    CwlBooleanParameterModel,
    CwlDirectoryParameterModel,
    CwlFileParameterModel,
    CwlFloatParameterModel,
    CwlIntegerParameterModel,
    CwlNullParameterModel,
    CwlStringParameterModel,
    CwlUnionParameterModel,
    DataCollectionParameterModel,
    DataColumnParameterModel,
    DataParameterModel,
    DirectoryUriParameterModel,
    DrillDownParameterModel,
    FloatParameterModel,
    GenomeBuildParameterModel,
    GroupTagParameterModel,
    HiddenParameterModel,
    IntegerParameterModel,
    LabelValue,
    NumberCompatiableValidators,
    RepeatParameterModel,
    RulesParameterModel,
    SectionParameterModel,
    SelectCompatiableValidators,
    SelectParameterModel,
    TextCompatiableValidators,
    TextParameterModel,
    ToolParameterBundle,
    ToolParameterBundleModel,
    ToolParameterT,
)
from galaxy.util import string_as_bool


class ParameterDefinitionError(Exception):
    pass


class UnknownParameterTypeError(ParameterDefinitionError):
    pass


def get_color_value(input_source: InputSource) -> str:
    return input_source.get("value", "#000000")


def _from_input_source_galaxy(input_source: InputSource, profile: float) -> ToolParameterT:
    input_type = input_source.parse_input_type()
    if input_type == "param":
        param_type = input_source.get("type")
        if param_type == "integer":
            optional = input_source.parse_optional()
            value = input_source.get("value")
            int_value: Optional[int]
            if value:
                int_value = int(value)
            elif optional:
                int_value = None
            elif value == "" or value is None:
                # A truly required parameter: https://github.com/galaxyproject/galaxy/pull/16966/files
                int_value = None
            else:
                raise ParameterDefinitionError()
            static_validator_models = static_validators(input_source.parse_validators())
            int_validators: List[NumberCompatiableValidators] = []
            for static_validator in static_validator_models:
                if static_validator.type == "in_range":
                    int_validators.append(static_validator)
            min_raw = input_source.get("min", None)
            max_raw = input_source.get("max", None)
            min_int = int(min_raw) if min_raw is not None else None
            max_int = int(max_raw) if max_raw is not None else None
            return IntegerParameterModel(
                type="integer",
                name=input_source.parse_name(),
                optional=optional,
                value=int_value,
                min=min_int,
                max=max_int,
                validators=int_validators,
            )
        elif param_type == "boolean":
            nullable = input_source.parse_optional()
            value = input_source.get_bool_or_none("checked", None if nullable else False)
            return BooleanParameterModel(
                type="boolean",
                name=input_source.parse_name(),
                optional=nullable,
                value=value,
            )
        elif param_type == "text":
            optional, optionality_inferred = text_input_is_optional(input_source)
            implicit_default = None if optional else ""
            default_value = input_source.get("value", implicit_default)
            text_validators: List[TextCompatiableValidators] = _text_validators(input_source)
            return TextParameterModel(
                type="text",
                name=input_source.parse_name(),
                optional=optional,
                validators=text_validators,
                value=default_value,
            )
        elif param_type == "float":
            optional = input_source.parse_optional()
            value = input_source.get("value")
            float_value: Optional[float]
            if value:
                float_value = float(value)
            elif optional:
                float_value = None
            elif value == "" or value is None:
                # See note above for matching integer handling.
                float_value = None
            else:
                raise ParameterDefinitionError()
            static_validator_models = static_validators(input_source.parse_validators())
            float_validators: List[NumberCompatiableValidators] = []
            for static_validator in static_validator_models:
                if static_validator.type == "in_range":
                    float_validators.append(static_validator)
            min_raw = input_source.get("min", None)
            max_raw = input_source.get("max", None)
            min_float = float(min_raw) if min_raw is not None else None
            max_float = float(max_raw) if max_raw is not None else None
            return FloatParameterModel(
                type="float",
                name=input_source.parse_name(),
                optional=optional,
                value=float_value,
                min=min_float,
                max=max_float,
                validators=float_validators,
            )
        elif param_type == "hidden":
            optional = input_source.parse_optional()
            value = input_source.get("value")
            hidden_validators: List[TextCompatiableValidators] = _text_validators(input_source)
            return HiddenParameterModel(
                type="hidden",
                name=input_source.parse_name(),
                optional=optional,
                value=value,
                validators=hidden_validators,
            )
        elif param_type == "color":
            optional = input_source.parse_optional()
            return ColorParameterModel(
                type="color",
                name=input_source.parse_name(),
                optional=optional,
                value=get_color_value(input_source),
            )
        elif param_type == "rules":
            return RulesParameterModel(
                type="rules",
                name=input_source.parse_name(),
            )
        elif param_type in ("data", "hidden_data"):
            # hidden_data is broken without optional="true" (job runner rejects
            # missing datasets); only known user is cufflinks which sets it.
            optional = input_source.parse_optional() if param_type == "data" else True
            multiple = input_source.get_bool("multiple", False)
            return DataParameterModel(
                type="data",
                name=input_source.parse_name(),
                optional=optional,
                multiple=multiple,
            )
        elif param_type == "data_collection":
            optional = input_source.parse_optional()
            default_value = input_source.parse_default()
            return DataCollectionParameterModel(
                collection_type=input_source.get("collection_type"),
                type="data_collection",
                name=input_source.parse_name(),
                optional=optional,
                value=default_value,
            )
        elif param_type == "select":
            # Function... example in devteam cummeRbund.
            optional = input_source.parse_optional()
            dynamic_options_config = input_source.parse_dynamic_options()
            is_static = dynamic_options_config is None
            multiple = input_source.get_bool("multiple", False)
            options: Optional[List[LabelValue]] = None
            if is_static:
                options = []
                for option_label, option_value, selected in input_source.parse_static_options():
                    options.append(LabelValue(label=option_label, value=option_value, selected=selected))
            static_validator_models = static_validators(input_source.parse_validators())
            select_validators: List[SelectCompatiableValidators] = []
            for static_validator in static_validator_models:
                if static_validator.type == "no_options":
                    # test case test_tool_execute::test_select_optional_null_by_default verifies
                    # these validators don't get applied effectively if the select is optional.
                    if not optional:
                        select_validators.append(static_validator)
            return SelectParameterModel(
                type="select",
                name=input_source.parse_name(),
                optional=optional,
                options=options,
                multiple=multiple,
                validators=select_validators,
            )
        elif param_type == "drill_down":
            multiple = input_source.get_bool("multiple", False)
            hierarchy = input_source.get("hierarchy", "exact")
            dynamic_options = input_source.parse_drill_down_dynamic_options()
            static_options = None
            if dynamic_options is None:
                static_options = input_source.parse_drill_down_static_options()
            return DrillDownParameterModel(
                type="drill_down",
                name=input_source.parse_name(),
                multiple=multiple,
                hierarchy=hierarchy,
                options=static_options,
            )
        elif param_type == "data_column":
            multiple = input_source.get_bool("multiple", False)
            optional = input_source.parse_optional()
            value = input_source.get("value")
            # mirror basic.py logic around accept_default
            accept_default = input_source.get_bool("accept_default", False)
            if not optional and accept_default:
                optional = True
            if accept_default and value is None:
                value = [0] if multiple else 0

            if isinstance(value, str):
                if multiple:
                    value = [int(v) for v in multiple_select_value_split(value)]
                else:
                    value = int(value)
            return DataColumnParameterModel(
                type="data_column",
                name=input_source.parse_name(),
                multiple=multiple,
                optional=optional,
                value=value,
            )
        elif param_type == "group_tag":
            multiple = input_source.get_bool("multiple", False)
            optional = input_source.parse_optional()
            return GroupTagParameterModel(
                type="group_tag",
                name=input_source.parse_name(),
                optional=optional,
                multiple=multiple,
            )
        elif param_type == "baseurl":
            return BaseUrlParameterModel(
                type="baseurl",
                name=input_source.parse_name(),
            )
        elif param_type == "genomebuild":
            optional = input_source.parse_optional()
            multiple = input_source.get_bool("multiple", False)
            return GenomeBuildParameterModel(
                type="genomebuild",
                name=input_source.parse_name(),
                optional=optional,
                multiple=multiple,
            )
        elif param_type == "directory_uri":
            directory_uri_validators: List[TextCompatiableValidators] = _text_validators(input_source)
            return DirectoryUriParameterModel(
                type="directory",
                name=input_source.parse_name(),
                validators=directory_uri_validators,
            )
        else:
            raise UnknownParameterTypeError(f"Unknown Galaxy parameter type {param_type}")
    elif input_type == "conditional":
        test_param_input_source = input_source.parse_test_input_source()
        test_parameter = cast(
            Union[BooleanParameterModel, SelectParameterModel],
            _from_input_source_galaxy(test_param_input_source, profile),
        )
        whens = []
        default_test_value = cond_test_parameter_default_value(test_parameter)
        for value, case_inputs_sources in input_source.parse_when_input_sources():
            if isinstance(test_parameter, BooleanParameterModel):
                true_value = test_param_input_source.get("truevalue")
                false_value = test_param_input_source.get("falsevalue")
                if isinstance(value, str) and value == true_value:
                    value = True
                elif isinstance(value, str) and value == false_value:
                    value = False
                typed_value = string_as_bool(value)
            else:
                typed_value = value

            tool_parameter_models = input_models_for_page(case_inputs_sources, profile)
            is_default_when = False
            if typed_value == default_test_value:
                is_default_when = True
            whens.append(
                ConditionalWhen(
                    discriminator=typed_value, parameters=tool_parameter_models, is_default_when=is_default_when
                )
            )
        return ConditionalParameterModel(
            type="conditional",
            name=input_source.parse_name(),
            test_parameter=test_parameter,
            whens=whens,
        )
    elif input_type == "repeat":
        name = input_source.get("name")
        # title = input_source.get("title")
        # help = input_source.get("help", None)
        instance_sources = input_source.parse_nested_inputs_source()
        instance_tool_parameter_models = input_models_for_page(instance_sources, profile)
        min_raw = input_source.get("min", None)
        max_raw = input_source.get("max", None)
        min = int(min_raw) if min_raw is not None else None
        max = int(max_raw) if max_raw is not None else None
        return RepeatParameterModel(
            type="repeat",
            name=name,
            parameters=instance_tool_parameter_models,
            min=min,
            max=max,
        )
    elif input_type == "section":
        name = input_source.get("name")
        instance_sources = input_source.parse_nested_inputs_source()
        instance_tool_parameter_models = input_models_for_page(instance_sources, profile)
        return SectionParameterModel(
            type="section",
            name=name,
            parameters=instance_tool_parameter_models,
        )
    else:
        raise Exception(
            f"Cannot generate tool parameter model for supplied tool source - unknown input_type {input_type}"
        )


def _simple_cwl_type_to_model(simple_type: str, input_source: CwlInputSource):
    if simple_type == "int":
        return CwlIntegerParameterModel(
            name=input_source.parse_name(),
        )
    elif simple_type == "float":
        return CwlFloatParameterModel(
            name=input_source.parse_name(),
        )
    elif simple_type == "null":
        return CwlNullParameterModel(
            name=input_source.parse_name(),
        )
    elif simple_type == "string":
        return CwlStringParameterModel(
            name=input_source.parse_name(),
        )
    elif simple_type == "boolean":
        return CwlBooleanParameterModel(
            name=input_source.parse_name(),
        )
    elif simple_type == "org.w3id.cwl.cwl.File":
        return CwlFileParameterModel(
            name=input_source.parse_name(),
        )
    elif simple_type == "org.w3id.cwl.cwl.Directory":
        return CwlDirectoryParameterModel(
            name=input_source.parse_name(),
        )
    raise NotImplementedError(
        f"Cannot generate tool parameter model for this CWL artifact yet - contains unknown type {simple_type}."
    )


def _text_validators(input_source: InputSource) -> List[TextCompatiableValidators]:
    static_validator_models = static_validators(input_source.parse_validators())
    text_validators: List[TextCompatiableValidators] = []
    for static_validator in static_validator_models:
        if static_validator.type == "length":
            text_validators.append(static_validator)
        elif static_validator.type == "regex":
            text_validators.append(static_validator)
        elif static_validator.type == "expression":
            text_validators.append(static_validator)
        elif static_validator.type == "empty_field":
            text_validators.append(static_validator)
    return text_validators


def _from_input_source_cwl(input_source: CwlInputSource) -> ToolParameterT:
    schema_salad_field = input_source.field
    if schema_salad_field is None:
        raise NotImplementedError("Cannot generate tool parameter model for this CWL artifact yet.")
    if "type" not in schema_salad_field:
        raise NotImplementedError("Cannot generate tool parameter model for this CWL artifact yet.")
    schema_salad_type = schema_salad_field["type"]
    if isinstance(schema_salad_type, str):
        return _simple_cwl_type_to_model(schema_salad_type, input_source)
    elif isinstance(schema_salad_type, list):
        return CwlUnionParameterModel(
            name=input_source.parse_name(),
            parameters=[_simple_cwl_type_to_model(t, input_source) for t in schema_salad_type],
        )
    else:
        raise NotImplementedError("Cannot generate tool parameter model for this CWL artifact yet.")


def input_models_from_json(json: List[Dict[str, Any]]) -> ToolParameterBundle:
    return ToolParameterBundleModel(parameters=json)


def tool_parameter_bundle_from_json(json: Dict[str, Any]) -> ToolParameterBundleModel:
    return ToolParameterBundleModel(**json)


def input_models_for_tool_source(tool_source: ToolSource) -> ToolParameterBundleModel:
    pages = tool_source.parse_input_pages()
    profile = parse_profile_version(tool_source)
    return ToolParameterBundleModel(parameters=input_models_for_pages(pages, profile))


def input_models_for_pages(pages: PagesSource, profile: float) -> List[ToolParameterT]:
    input_models = []
    if pages.inputs_defined:
        for page_source in pages.page_sources:
            input_models.extend(input_models_for_page(page_source, profile))

    return input_models


def input_models_for_page(page_source: PageSource, profile: float) -> List[ToolParameterT]:
    input_models = []
    for input_source in page_source.parse_input_sources():
        input_type = input_source.parse_input_type()
        if input_type == "display":
            # not a real input... just skip this. Should this be handled in the parser layer better?
            continue
        tool_parameter_model = from_input_source(input_source, profile)
        input_models.append(tool_parameter_model)
    return input_models


def from_input_source(input_source: InputSource, profile: float) -> ToolParameterT:
    tool_parameter: ToolParameterT
    if isinstance(input_source, CwlInputSource):
        tool_parameter = _from_input_source_cwl(input_source)
    else:
        tool_parameter = _from_input_source_galaxy(input_source, profile)
    return tool_parameter



from urllib.parse import quote


def uri_safe_tool_id(tool_id: str) -> str:
    # Shed tool ids contain ``+`` in their version suffix (e.g.
    # ``.../1.3.1+galaxy1``). In a query string the ``+`` decodes to a
    # literal space unless percent-encoded as ``%2B``.
    return quote(tool_id)



"""This module contains citation linters.

Citations describe references that should be used when consumers
of the tool publish results.
"""

from typing import TYPE_CHECKING

from galaxy.tool_util.lint import Linter

if TYPE_CHECKING:
    from galaxy.tool_util.lint import LintContext
    from galaxy.tool_util.parser.interface import ToolSource


class CitationsMissing(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        root = tool_xml.find("./citations")
        if root is None:
            root = tool_xml.getroot()
        citations = tool_xml.findall("citations")
        if len(citations) == 0:
            lint_ctx.warn("No citations found, consider adding citations to your tool.", linter=cls.name(), node=root)


class CitationsNoText(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        citations = tool_xml.find("citations")
        if citations is None:
            return
        for citation in citations:
            citation_type = citation.attrib.get("type")
            if citation_type in ["doi", "bibtex"] and (citation.text is None or not citation.text.strip()):
                lint_ctx.error(f"Empty {citation_type} citation.", linter=cls.name(), node=citation)


class CitationsFound(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        root = tool_xml.find("./citations")
        if root is None:
            root = tool_xml.getroot()
        citations = tool_xml.find("citations")

        if citations is not None and len(citations) > 0:
            lint_ctx.valid(f"Found {len(citations)} citations.", linter=cls.name(), node=root)


class CitationsNoValid(Linter):
    @classmethod
    def lint(cls, tool_source: "ToolSource", lint_ctx: "LintContext"):
        tool_xml = getattr(tool_source, "xml_tree", None)
        if not tool_xml:
            return
        root = tool_xml.find("./citations")
        if root is None:
            root = tool_xml.getroot()
        citations = tool_xml.findall("citations")
        if len(citations) != 1:
            return
        if len(citations[0]) == 0:
            lint_ctx.warn("Found no valid citations.", linter=cls.name(), node=root)
